"""Helper functions for TellorX contracts."""
import logging
from typing import Any
from typing import Optional
from typing import Tuple

from telliot_core.contract.contract import Contract
from telliot_core.utils.response import error_status
from telliot_core.utils.response import ResponseStatus
from web3.datastructures import AttributeDict

logger = logging.getLogger(__name__)


async def write_with_retry(
    contract: Contract,
    func_name: str,
    extra_gas_price: int,
    retries: int,
    gas_limit: int,
    legacy_gas_price: Optional[int] = None,
    max_priority_fee_per_gas: Optional[int] = None,
    max_fee_per_gas: Optional[int] = None,
    **kwargs: Any,
) -> Tuple[Optional[AttributeDict[Any, Any]], ResponseStatus]:
    """For submitting any contract transaction.
    Will attempt to retry (resubmit) the transaction a set number of times
    if it fails to be picked up by the chain.
    Can submit a tx with either legacy gas strategy
    or the London fork EIP-1559 strategy

    Quick explainer...
    Legacy: gas cost = gasLimit*gasPrice where...
     - you set the gas price (simple as that!)
    EIP-1559: gas cost = gasLimit*(baseFee+priorityFee) where...
     - you can set the highest priorityFee you would pay (max_priority_fee_per_gas)
     OR
     - you can set the highest maxFee (i.e. baseFee+priorityFee)
       you are willing to pay (ths is the max_fee_per_gas)
    also...
     - baseFee is set by the block

    Args:
        contract: Contract object
        func_name (str): name of contract function
        extra_gas_price (int): the amount of gas the tx resends
            with if it fails due to gas strategy
            (adds to legacy_gas_price or max_priority_fee_per_gas
            depending on tx gas strategy)
        retries (int): number of times to attempt tx resubmission
        gas_limit (int): the maximum amount of gas units (not gas price!)
            to assign to the tx
        legacy_gas_price (int): the legacy gas price for a legacy tx
        max_priority_fee_per_gas (int): the highest priorityFee in gwei
            you would pay (see guide above)
        max_fee_per_gas (int): the highest maxFee in gwei
            (baseFee+priorityFee) you would pay (see guide above)

    Returns:
        transaction receipt (if transaction is picked up by chain)
        ResponseStatus (status of tx success)


    """

    try:
        status = ResponseStatus()
        acc = contract.node.web3.eth.account.from_key(contract.private_key)
        acc_nonce = contract.node.web3.eth.get_transaction_count(acc.address)

        # Iterate through retry attempts
        for k in range(retries + 1):

            attempt = k + 1

            if k >= 1:
                logger.info(f"Retrying {func_name} (attempt #{attempt})")

            # Attempt write
            tx_receipt, status = await contract.write(
                func_name=func_name,
                legacy_gas_price=legacy_gas_price,
                max_priority_fee_per_gas=max_priority_fee_per_gas,
                max_fee_per_gas=max_fee_per_gas,
                acc_nonce=acc_nonce,
                gas_limit=gas_limit,
                **kwargs,
            )

            logger.debug(f"Attempt {attempt} status: {status}")

            # Exit loop if transaction successful
            if status.ok and (tx_receipt is not None) and (tx_receipt["status"] == 1):
                return tx_receipt, status

            else:
                logger.info(f"Write attempt {attempt} failed:")
                msg = str(status.error)
                _ = error_status(msg, log=logger.info)

                if tx_receipt is not None:
                    tx_url = f"{contract.node.explorer}/tx/{tx_receipt['transactionHash'].hex()}"  # noqa: E501

                    if tx_receipt["status"] == 0:
                        msg = f"Write attempt {attempt} failed, tx reverted ({tx_url}):"  # noqa: E501
                        return tx_receipt, error_status(msg, log=logger.info)

                    else:
                        msg = f"Write attempt {attempt}: Invalid TX Receipt status: {tx_receipt['status']}"  # noqa: E501
                        return tx_receipt, error_status(msg, log=logger.info)

                if status.error:
                    if "replacement transaction underpriced" in status.error:
                        if legacy_gas_price is not None:
                            legacy_gas_price += extra_gas_price
                            logger.info(f"Next gas price: {legacy_gas_price}")
                        elif max_fee_per_gas is not None:
                            max_fee_per_gas += extra_gas_price
                            logger.info(f"Next max fee: {max_fee_per_gas}")
                        elif max_priority_fee_per_gas is not None:
                            max_priority_fee_per_gas += extra_gas_price
                            logger.info(
                                f"Next priority fee: {max_priority_fee_per_gas}"
                            )
                    elif "already known" in status.error:
                        acc_nonce += 1
                        logger.info(f"Incrementing nonce: {acc_nonce}")
                    elif "nonce too low" in status.error:
                        acc_nonce += 1
                        logger.info(f"Incrementing nonce: {acc_nonce}")
                    # a different rpc error
                    elif "nonce is too low" in status.error:
                        acc_nonce += 1
                        logger.info(f"Incrementing nonce: {acc_nonce}")
                    elif "not in the chain" in status.error:
                        if legacy_gas_price is not None:
                            legacy_gas_price += extra_gas_price
                            logger.info(f"Next gas price: {legacy_gas_price}")
                        elif max_fee_per_gas is not None:
                            max_fee_per_gas += extra_gas_price
                            logger.info(f"Next max fee: {max_fee_per_gas}")
                        elif max_priority_fee_per_gas is not None:
                            max_priority_fee_per_gas += extra_gas_price
                            logger.info(
                                f"Next priority fee: {max_priority_fee_per_gas}"
                            )
                    else:
                        extra_gas_price = 0

        status.ok = False
        status.error = "ran out of retries, tx unsuccessful"

        return tx_receipt, status

    except Exception as e:
        return None, error_status("Other error", log=logger.error, e=e)



from dataclasses import dataclass
from datetime import datetime
from typing import Any
from typing import Optional
from typing import Tuple
from urllib.parse import urlencode

import requests
from telliot_core.dtypes.datapoint import datetime_now_utc
from telliot_core.dtypes.datapoint import OptionalDataPoint
from telliot_core.pricing.price_service import WebPriceService
from telliot_core.pricing.price_source import PriceSource

from telliot_feed_examples.utils.log import get_logger


logger = get_logger(__name__)


# Hardcoded supported assets & currencies
# Kraken uses XBT instead of BTC for its APIs:
# https://support.kraken.com/hc/en-us/articles/360001206766-Bitcoin-currency-code-XBT-vs-BTC
kraken_assets = {"ETH", "XBT"}
kraken_currencies = {"USD"}


class KrakenHistoricalPriceService(WebPriceService):
    """Kraken Historical Price Service"""

    def __init__(
        self,
        timeout: float = 0.5,
        name: str = "Kraken Historical Price Service",
        url: str = "https://api.kraken.com",
        ts: int = 0,
    ):
        self.name = name
        self.url = url
        self.ts = ts
        self.timeout = timeout

    def get_url(self, url: str = "") -> dict[str, Any]:
        """Helper function to get URL JSON response while handling exceptions

        Args:
            url: URL to fetch

        Returns:
            A dictionary with the following (optional) keys:
                json (dict or list): Result, if no error occurred
                error (str): A description of the error, if one occurred
                exception (Exception): The exception, if one occurred
        """

        request_url = self.url + url

        with requests.Session() as s:
            try:
                r = s.get(request_url, timeout=self.timeout)
                json_data = r.json()
                return {"response": json_data}

            except requests.exceptions.ConnectTimeout as e:
                return {"error": "Timeout Error", "exception": e}

            except Exception as e:
                return {"error": str(type(e)), "exception": e}

    def get_request_url(self, asset: str, currency: str, period_start: int) -> str:
        """Assemble Kraken historical trades request url."""
        asset = asset.upper()
        currency = currency.upper()

        if asset not in kraken_assets:
            raise Exception(f"Asset not supported: {asset}")
        if currency not in kraken_currencies:
            raise Exception(f"Currency not supported: {currency}")

        url_params = urlencode(
            {"pair": f"{asset}{currency}", "since": period_start}  # Unix timestamp
        )

        # Source: https://docs.kraken.com/rest/#operation/getRecentTrades
        return f"/0/public/Trades?{url_params}"

    def resp_price_parse(
        self, asset: str, currency: str, resp: dict[Any, Any]
    ) -> Optional[float]:
        """Gets first price from trades data."""
        try:
            # Price of last trade in trades list retrieved from API
            pair_key = f"X{asset.upper()}Z{currency.upper()}"
            trades = resp["result"][pair_key]
        except KeyError as e:
            msg = f"Error parsing Kraken API response: KeyError: {e}"
            logger.critical(msg)
            return None

        if len(trades) == 0:
            logger.warning("No trades found.")
            return None

        return float(trades[-1][0])

    def resp_all_trades_parse(
        self, asset: str, currency: str, resp: dict[Any, Any]
    ) -> Optional[list[str]]:
        """Gets all trades."""
        trades = None
        try:
            pair_key = f"X{asset.upper()}Z{currency.upper()}"
            trades = resp["result"][pair_key]
        except KeyError as e:
            msg = f"Error parsing Kraken API response: KeyError: {e}"
            logger.critical(msg)

        return trades

    async def get_price(
        self, asset: str, currency: str, ts: Optional[int] = None
    ) -> OptionalDataPoint[float]:
        """Implement PriceServiceInterface

        This implementation gets the historical price from
        the Kraken API using a timestamp."""
        if ts is None:
            ts = self.ts

        req_url = self.get_request_url(asset, currency, ts)

        d = self.get_url(req_url)

        if "error" in d:
            logger.error(d)
            return None, None

        elif "response" in d:
            response = d["response"]
            price = self.resp_price_parse(resp=response, asset=asset, currency=currency)

        else:
            raise Exception("Invalid response from get_url")

        return price, datetime_now_utc()

    async def get_trades(
        self,
        asset: str,
        currency: str,
        period: int = 900,  # 15 minutes
        ts: Optional[int] = None,
    ) -> Tuple[Optional[list[str]], Optional[datetime]]:
        """Retrieves list of historical prices.

        Gets historical prices of trades data from Kraken API using a timestamp.
        Returns all trades data within the given period. The period is moved backwards
        in time if current time overlaps.

        Args:
            period: time window in seconds of retrieved trades
            timestamp: unix timestamp, beginning of
            the period (time window of fetched trades data)

        Returns:
            trades: list of trades (sublists). The elements of a trade
            sublist (in order) are price, volume, time, buy/sell, market/limit,
            and miscellaneous."""

        if ts is None:
            ts = self.ts

        period_start = int(ts - period)

        req_url = self.get_request_url(asset, currency, period_start)

        d = self.get_url(req_url)

        if "error" in d:
            logger.error(d)
            return None, None

        elif "response" in d:
            response = d["response"]
            trades = self.resp_all_trades_parse(
                resp=response, asset=asset, currency=currency
            )

        else:
            raise Exception("Invalid response from get_url")

        return trades, datetime_now_utc()


@dataclass
class KrakenHistoricalPriceSource(PriceSource):
    ts: int = 0
    asset: str = ""
    currency: str = ""
    service: KrakenHistoricalPriceService = KrakenHistoricalPriceService(ts=ts)

    def __post_init__(self) -> None:
        self.service.ts = self.ts



from dataclasses import dataclass
from dataclasses import field
from typing import Any

from telliot_core.dtypes.datapoint import datetime_now_utc
from telliot_core.dtypes.datapoint import OptionalDataPoint
from telliot_core.pricing.price_service import WebPriceService
from telliot_core.pricing.price_source import PriceSource

from telliot_feed_examples.utils.log import get_logger


logger = get_logger(__name__)
pancakeswap_map = {
    "wbnb": "0xbb4cdb9cbd36b01bd1cbaebf2de08d9173bc095c",
    "fuse": "0x5857c96dae9cf8511b08cb07f85753c472d36ea3",
}


class PancakeswapPriceService(WebPriceService):
    """Pancakeswap Price Service in USD and BNB"""

    def __init__(self, **kwargs: Any) -> None:
        kwargs["name"] = "Pancakeswap Price Service"
        kwargs["url"] = "https://api.pancakeswap.info"
        super().__init__(**kwargs)

    async def get_price(self, asset: str, currency: str) -> OptionalDataPoint[float]:
        """Implement PriceServiceInterface

        This implementation gets the price from the Pancakeswap API
        Pankcakeswap official Github repo for API:
        https://github.com/pancakeswap/pancake-info-api

        """

        asset = asset.lower()

        token_addr = pancakeswap_map.get(asset, None)

        if not token_addr:
            raise Exception("Asset not supported: {}".format(asset))

        request_url = f"/api/v2/tokens/{token_addr}"

        d = self.get_url(request_url)

        if "error" in d:
            logger.error(d)
            return None, None

        elif "response" in d:
            response = d["response"]
            print(response)

            try:
                price = (
                    response["data"]["price"]
                    if currency.lower() == "usd"
                    else response["data"]["price_BNB"]
                )
            except KeyError as e:
                msg = "Error parsing Pancakeswap API response: KeyError: {}".format(e)
                logger.critical(msg)

        else:
            raise Exception("Invalid response from get_url")
        logger.info(f"price for {asset} in {currency}: {price}")
        return float(price), datetime_now_utc()


@dataclass
class PancakeswapPriceSource(PriceSource):
    asset: str = ""
    currency: str = ""
    service: PancakeswapPriceService = field(
        default_factory=PancakeswapPriceService, init=False
    )



# Majority of this code from web3-flashbots:
# https://github.com/flashbots/web3-flashbots
# Including it here instead because it ^ doesn't seem to be actively supported.
# EIP-1559 subbport by @lekhovitsky
# https://github.com/lekhovitsky
# type: ignore
from typing import Optional
from typing import Union

from eth_account.signers.local import LocalAccount
from eth_typing import URI
from web3 import Web3
from web3._utils.module import attach_modules

from .flashbots import Flashbots
from .middleware import construct_flashbots_middleware
from .provider import FlashbotProvider


DEFAULT_FLASHBOTS_RELAY = "https://relay.flashbots.net"


def flashbot(
    w3: Web3,
    signature_account: LocalAccount,
    endpoint_uri: Optional[Union[URI, str]] = None,
):
    """
    Injects the flashbots module and middleware to w3.
    """

    flashbots_provider = FlashbotProvider(signature_account, endpoint_uri)
    flash_middleware = construct_flashbots_middleware(flashbots_provider)
    w3.middleware_onion.add(flash_middleware)

    # attach modules to add the new namespace commands
    attach_modules(w3, {"flashbots": (Flashbots,)})



import os
from datetime import datetime

import pytest
from telliot_core.apps.telliot_config import TelliotConfig

from telliot_feed_examples.sources.ampl_usd_vwap import AMPLUSDVWAPSource
from telliot_feed_examples.sources.ampl_usd_vwap import AnyBlockSource
from telliot_feed_examples.sources.ampl_usd_vwap import BraveNewCoinSource


@pytest.fixture()
def keys_dict():
    keys = TelliotConfig().api_keys
    anyblock_key = keys.find("anyblock")[0].key
    rapid_key = keys.find("bravenewcoin")[0].key

    if not anyblock_key and "ANYBLOCK_KEY" in os.environ:
        anyblock_key = os.environ["ANYBLOCK_KEY"]

    if not rapid_key and "RAPID_KEY" in os.environ:
        rapid_key = os.environ["RAPID_KEY"]

    return {"anyblock": anyblock_key, "bravenewcoin": rapid_key}


@pytest.mark.asyncio
async def test_bravenewcoin_source(keys_dict):
    """Test retrieving AMPL/USD/VWAP data from BraveNewCoin/Rapid api.

    Retrieves bearer token and adds to headers of main data request."""

    if keys_dict["bravenewcoin"]:
        ampl_source = BraveNewCoinSource(api_key=keys_dict["bravenewcoin"])

        value, timestamp = await ampl_source.fetch_new_datapoint()

        assert isinstance(value, float)
        assert isinstance(timestamp, datetime)
        assert value > 0
    else:
        print("No BraveNewCoin api key ")


@pytest.mark.asyncio
async def test_anyblock_source(keys_dict):
    """Test retrieving AMPL/USD/VWAP data from AnyBlock api."""
    api_key = keys_dict["anyblock"]

    if api_key != "":
        ampl_source = AnyBlockSource(api_key=api_key)

        value, timestamp = await ampl_source.fetch_new_datapoint()

        assert isinstance(value, float)
        assert isinstance(timestamp, datetime)
        assert value > 0
    else:
        print("No AnyBlock api key ")


@pytest.mark.asyncio
async def test_ampl_usd_vwap_source(keys_dict):
    """Test getting median updated AMPL/USD/VWAP value."""

    if not all(keys_dict.values()):
        print("No api keys found")
    else:
        ampl_source = AMPLUSDVWAPSource()
        ampl_source.sources[0].api_key = keys_dict["anyblock"]
        ampl_source.sources[1].api_key = keys_dict["bravenewcoin"]

        value, timestamp = await ampl_source.fetch_new_datapoint()

        assert isinstance(value, float)
        assert isinstance(timestamp, datetime)
        assert value > 0


@pytest.mark.asyncio
async def test_no_updated_value(bad_source):
    """Test no AMPL/USD/VWAP value retrieved."""

    ampl_source = AMPLUSDVWAPSource()

    # Switch source to test one that doesn't return an updated value
    ampl_source.sources = [bad_source]

    value, timestamp = await ampl_source.fetch_new_datapoint()

    assert value is None
    assert timestamp is None



import asyncio
import os

from abi import erc20_abi
from abi import tellor_flex_abi
from dotenv import find_dotenv
from dotenv import load_dotenv
from eth_account.account import Account
from eth_account.signers.local import LocalAccount
from web3 import HTTPProvider
from web3 import Web3
from web3.middleware import geth_poa_middleware

from telliot_feed_examples.feeds.trb_usd_feed import trb_usd_median_feed


load_dotenv(find_dotenv())


def env(key: str) -> str:
    return os.environ.get(key)


async def main() -> None:
    account: LocalAccount = Account.from_key(env("polygon_pk"))

    chain_id = 80001
    staking_amount = int(1e18 * 10)  # 10 trb
    print("staking amount:", staking_amount)

    w3 = Web3(HTTPProvider(env("polygon_endpoint_url")))
    if chain_id == 80001:
        w3.middleware_onion.inject(geth_poa_middleware, layer=0)

    print("address:", account.address)
    print(
        f"""Reporter account balance:
        {Web3.fromWei(w3.eth.get_balance(account.address), 'ether')} MATIC
        """
    )

    # Mainnet
    if chain_id == 137:
        flex_address = "0xFd45Ae72E81Adaaf01cC61c8bCe016b7060DD537"
        trb_address = "0xE3322702BEdaaEd36CdDAb233360B939775ae5f1"
    # Mumbai testnet
    else:
        flex_address = "0x41b66dd93b03e89D29114a7613A6f9f0d4F40178"
        trb_address = "0x45cAF1aae42BA5565EC92362896cc8e0d55a2126"

    oracle = w3.eth.contract(
        address=flex_address,
        abi=tellor_flex_abi,
    )

    trb_polygon = w3.eth.contract(
        address=trb_address,
        abi=erc20_abi,
    )

    staker_info_func = oracle.get_function_by_name("getStakerInfo")
    staker_info = staker_info_func(_staker=account.address).call()
    (
        staker_startdate,
        staked_balance,
        locked_balance,
        last_report,
        num_reports,
    ) = staker_info
    print("staker start date:", staker_startdate)
    print("staked balance:", staked_balance)
    print("locked balance:", locked_balance)
    print("last report:", last_report)
    print("num reports:", num_reports)

    # # Transfer TRB & MATIC to dev-acct-9
    # print('testing transfer')
    # transfer_func = trb_polygon.get_function_by_name("transfer")
    # acc_nonce = w3.eth.get_transaction_count(account.address)
    # transfer_tx_dict = {
    #     "nonce": acc_nonce,
    #     "gas": 70000,
    #     "maxFeePerGas": Web3.toWei(1000, "gwei"),
    #     "maxPriorityFeePerGas": Web3.toWei(100, "gwei"),
    #     "chainId": chain_id,
    # }
    # transfer_func = transfer_func(
    #     recipient="0xE0Fe38e04b8952120a8015bE08248aC85891b477", # dev-acct-9
    #     amount=1)
    # built_transfer_func = transfer_func.buildTransaction(transfer_tx_dict)
    # tx_signed = account.sign_transaction(built_transfer_func)
    # tx_hash = w3.eth.send_raw_transaction(tx_signed.rawTransaction)

    # tx_receipt = w3.eth.wait_for_transaction_receipt(
    #     tx_hash, timeout=360
    # )
    # print(tx_receipt)

    # tx_url = f"https://polygonscan.com/tx/{tx_hash.hex()}"

    # print(f"view transfer transaction: ({tx_url})")

    if staker_startdate == 0:
        print()
        print("approving deposit stake")
        approve_func = trb_polygon.get_function_by_name("approve")
        acc_nonce = w3.eth.get_transaction_count(account.address)
        approve_tx_dict = {
            "nonce": acc_nonce,
            "gas": 100000,
            "maxFeePerGas": Web3.toWei(1000, "gwei"),
            "maxPriorityFeePerGas": Web3.toWei(100, "gwei"),
            "chainId": chain_id,
        }
        approve_func = approve_func(spender=flex_address, amount=staking_amount)
        built_approve_func = approve_func.buildTransaction(approve_tx_dict)
        tx_signed = account.sign_transaction(built_approve_func)
        tx_hash = w3.eth.send_raw_transaction(tx_signed.rawTransaction)

        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash, timeout=360)
        print(tx_receipt)

        tx_url = f"https://polygonscan.com/tx/{tx_hash.hex()}"

        print(f"view approve transaction: ({tx_url})")

        print("depositing stake")
        stake_func = oracle.get_function_by_name("depositStake")
        acc_nonce = w3.eth.get_transaction_count(account.address)
        stake_tx_dict = {
            "nonce": acc_nonce,
            "gas": 300000,
            "maxFeePerGas": Web3.toWei(1000, "gwei"),
            "maxPriorityFeePerGas": Web3.toWei(100, "gwei"),
            "chainId": chain_id,
        }
        stake_func = stake_func(_amount=staking_amount)
        built_stake_func = stake_func.buildTransaction(stake_tx_dict)
        tx_signed = account.sign_transaction(built_stake_func)
        tx_hash = w3.eth.send_raw_transaction(tx_signed.rawTransaction)

        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash, timeout=360)
        print(tx_receipt)

        tx_url = f"https://polygonscan.com/tx/{tx_hash.hex()}"

        print(f"view deposit stake transaction: ({tx_url})")

    _ = await trb_usd_median_feed.source.fetch_new_datapoint()
    latest_data = trb_usd_median_feed.source.latest
    if latest_data[0] is None:
        msg = "Unable to retrieve updated datafeed value."
        print(msg)

    query = trb_usd_median_feed.query
    nonce_getter = oracle.get_function_by_name("getNewValueCountbyQueryId")
    timestamp_count = nonce_getter(_queryId=query.query_id).call()
    print("timestamp count (nonce):", timestamp_count)

    submit_val_func = oracle.get_function_by_name("submitValue")
    submit_val_tx = submit_val_func(
        _queryId=query.query_id,
        _value=query.value_type.encode(latest_data[0]),
        _nonce=timestamp_count,
        _queryData=query.query_data,
    )
    acc_nonce = w3.eth.get_transaction_count(account.address)

    gas = 350000

    built_submit_val_tx = submit_val_tx.buildTransaction(
        {
            "nonce": acc_nonce,
            "gas": gas,
            "maxFeePerGas": Web3.toWei(1000, "gwei"),
            "maxPriorityFeePerGas": Web3.toWei(100, "gwei"),
            "chainId": chain_id,
        }
    )
    tx_signed = account.sign_transaction(built_submit_val_tx)

    tx_hash = w3.eth.send_raw_transaction(tx_signed.rawTransaction)

    tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash, timeout=360)
    print(tx_receipt)

    tx_url = f"https://polygonscan.com/tx/{tx_hash.hex()}"

    print(f"transaction succeeded. ({tx_url})")


if __name__ == "__main__":
    asyncio.run(main())



from dataclasses import dataclass

from telliot_core.datasource import DataSource
from telliot_core.dtypes.datapoint import DataPoint
from telliot_core.dtypes.datapoint import datetime_now_utc

from telliot_feed_examples.utils.log import get_logger


logger = get_logger(__name__)


@dataclass
class USPCESource(DataSource[float]):
    """DataSource for USPCE manually-entered data."""

    def parse_user_val(test_input: str) -> float:
        """Parse USPCE value from user input."""
        # This arg is to avoid a TypeError when the default
        # input() method is overriden in test_source.py.
        # The error says this method expects no params,
        # but is passed one. TODO: fix
        _ = test_input

        print("Enter USPCE value (example: 13659.3):")

        uspce = None

        while uspce is None:
            inpt = input()

            try:
                inpt = float(inpt)  # type: ignore
            except ValueError:
                print("Invalid input. Enter decimal value (float).")
                continue

            print(f"Submitting value: {inpt}\nPress [ENTER] to confirm.")
            _ = input()

            uspce = inpt

        return uspce

    async def fetch_new_datapoint(self) -> DataPoint[float]:
        """Update current value with time-stamped value fetched from user input.

        Returns:
            Current time-stamped value
        """
        uspce = self.parse_user_val()

        datapoint = (uspce, datetime_now_utc())
        self.store_datapoint(datapoint)

        logger.info(f"USPCE {datapoint[0]} retrieved at time {datapoint[1]}")

        return datapoint



"""Pytest Fixtures used for testing Pytelliot"""
import os

import pytest
from chained_accounts import ChainedAccount
from chained_accounts import find_accounts
from telliot_core.apps.telliot_config import TelliotConfig
from telliot_core.datasource import DataSource
from telliot_core.dtypes.datapoint import OptionalDataPoint


@pytest.fixture(scope="session", autouse=True)
def rinkeby_cfg():
    """Get rinkeby endpoint from config

    If environment variables are defined, they will override the values in config files
    """
    cfg = TelliotConfig()

    # Override configuration for rinkeby testnet
    cfg.main.chain_id = 4

    rinkeby_endpoint = cfg.get_endpoint()
    # assert rinkeby_endpoint.network == "rinkeby"

    if os.getenv("NODE_URL", None):
        rinkeby_endpoint.url = os.environ["NODE_URL"]

    rinkeby_accounts = find_accounts(chain_id=4)
    if not rinkeby_accounts:
        # Create a test account using PRIVATE_KEY defined on github.
        key = os.getenv("PRIVATE_KEY", None)
        if key:
            ChainedAccount.add(
                "git-rinkeby-key", chains=4, key=os.environ["PRIVATE_KEY"], password=""
            )
        else:
            raise Exception("Need a rinkeby account")

    return cfg


@pytest.fixture(scope="session", autouse=True)
def mumbai_cfg():
    """Return a test telliot configuration for use on polygon-mumbai

    If environment variables are defined, they will override the values in config files
    """
    cfg = TelliotConfig()

    # Override configuration for rinkeby testnet
    cfg.main.chain_id = 80001

    endpt = cfg.get_endpoint()
    if "INFURA_API_KEY" in endpt.url:
        endpt.url = (
            f'https://polygon-mumbai.infura.io/v3/{os.environ["INFURA_API_KEY"]}'
        )

    mumbai_accounts = find_accounts(chain_id=80001)
    if not mumbai_accounts:
        # Create a test account using PRIVATE_KEY defined on github.
        key = os.getenv("PRIVATE_KEY", None)
        if key:
            ChainedAccount.add(
                "git-mumbai-key",
                chains=80001,
                key=os.environ["PRIVATE_KEY"],
                password="",
            )
        else:
            raise Exception("Need a mumbai account")

    return cfg


@pytest.fixture(scope="session", autouse=True)
def ropsten_cfg():
    """Return a test telliot configuration for use on polygon-mumbai

    If environment variables are defined, they will override the values in config files
    """
    cfg = TelliotConfig()

    # Override configuration for ropsten testnet
    cfg.main.chain_id = 3

    endpt = cfg.get_endpoint()
    if "INFURA_API_KEY" in endpt.url:
        endpt.url = f'wss://ropsten.infura.io/ws/v3/{os.environ["INFURA_API_KEY"]}'

    accounts = find_accounts(chain_id=3)
    if not accounts:
        # Create a test account using PRIVATE_KEY defined on github.
        key = os.getenv("PRIVATE_KEY", None)
        if key:
            ChainedAccount.add(
                "git-ropsten-key", chains=3, key=os.environ["PRIVATE_KEY"], password=""
            )
        else:
            raise Exception("Need a ropsten account")

    return cfg


@pytest.fixture(scope="session")
def bad_source():
    """Used for testing no updated value for datafeeds."""

    class BadSource(DataSource[float]):
        """Source that does not return an updated DataPoint."""

        async def fetch_new_datapoint(self) -> OptionalDataPoint[float]:
            return None, None

    return BadSource()



from telliot_core.datafeed import DataFeed
from telliot_core.queries import SpotPrice

from telliot_feed_examples.sources.price.spot.coingecko import CoinGeckoSpotPriceSource
from telliot_feed_examples.sources.price_aggregator import PriceAggregator

ohm_eth_median_feed = DataFeed(
    query=SpotPrice(asset="OHM", currency="ETH"),
    source=PriceAggregator(
        asset="ohm",
        currency="eth",
        algorithm="median",
        sources=[CoinGeckoSpotPriceSource(asset="ohm", currency="eth")],
    ),
)



import pytest
from telliot_core.apps.core import TelliotCore
from web3.datastructures import AttributeDict

from telliot_feed_examples.feeds.uspce_feed import uspce_feed
from telliot_feed_examples.reporters.interval import IntervalReporter
from telliot_feed_examples.sources import uspce


@pytest.mark.skip("Asks for psswd")
@pytest.mark.asyncio
async def test_uspce_interval_reporter_submit_once(rinkeby_cfg):
    """test report of uspce manual price"""
    # Override Python built-in input method
    uspce.input = lambda: "123.456"

    async with TelliotCore(config=rinkeby_cfg) as core:

        account = core.get_account()
        tellorx = core.get_tellorx_contracts()
        r = IntervalReporter(
            endpoint=core.config.get_endpoint(),
            account=account,
            master=tellorx.master,
            oracle=tellorx.oracle,
            datafeed=uspce_feed,
            expected_profit="YOLO",
            transaction_type=0,
            gas_limit=400000,
            max_fee=None,
            priority_fee=None,
            legacy_gas_price=None,
            gas_price_speed="safeLow",
            chain_id=core.config.main.chain_id,
        )

        EXPECTED_ERRORS = {
            "Current addess disputed. Switch address to continue reporting.",
            "Current address is locked in dispute or for withdrawal.",
            "Current address is in reporter lock.",
            "Estimated profitability below threshold.",
            "Estimated gas price is above maximum gas price.",
            "Unable to retrieve updated datafeed value.",
        }

        ORACLE_ADDRESSES = {
            "0xe8218cACb0a5421BC6409e498d9f8CC8869945ea",  # mainnet
            "0x18431fd88adF138e8b979A7246eb58EA7126ea16",  # rinkeby
        }

        tx_receipt, status = await r.report_once()

        # Reporter submitted
        if tx_receipt is not None and status.ok:
            assert isinstance(tx_receipt, AttributeDict)
            assert tx_receipt.to in ORACLE_ADDRESSES
        # Reporter did not submit
        else:
            assert not tx_receipt
            assert not status.ok
            assert status.error in EXPECTED_ERRORS



""" Telliot Feed Examples CLI

A simple interface for interacting with telliot example feed functionality.
"""
import asyncio
import getpass
from typing import Optional
from typing import Union

import click
from chained_accounts import find_accounts
from click.core import Context
from eth_utils import to_checksum_address
from telliot_core.apps.core import TelliotCore
from telliot_core.cli.utils import async_run
from telliot_core.cli.utils import cli_core
from telliot_core.data.query_catalog import query_catalog
from telliot_core.tellor.tellorflex.diva import DivaOracleTellorContract

from telliot_feed_examples.feeds import CATALOG_FEEDS
from telliot_feed_examples.feeds.diva_protocol_feed import assemble_diva_datafeed
from telliot_feed_examples.reporters.flashbot import FlashbotsReporter
from telliot_feed_examples.reporters.interval import IntervalReporter
from telliot_feed_examples.reporters.tellorflex import PolygonReporter
from telliot_feed_examples.utils.log import get_logger
from telliot_feed_examples.utils.oracle_write import tip_query


logger = get_logger(__name__)

TELLOR_FLEX_CHAINS = (137, 80001, 3, 122)
DIVA_PROTOCOL_CHAINS = (137, 80001, 3)


def get_stake_amount() -> float:
    """Retrieve desired stake amount from user

    Each stake is 10 TRB on TellorFlex Polygon. If an address
    is not staked for any reason, the PolygonReporter will attempt
    to stake. Number of stakes determines the reporter lock:

    reporter_lock = 12hrs / N * stakes

    Retrieves desidred stake amount from user input."""

    msg = "Enter amount TRB to stake if unstaked:"
    stake = click.prompt(msg, type=float, default=10.0, show_default=True)
    assert isinstance(stake, float)
    assert stake >= 10.0

    return stake


def parse_profit_input(expected_profit: str) -> Optional[Union[str, float]]:
    """Parses user input expected profit and ensures
    the input is either a float or the string 'YOLO'."""
    if expected_profit == "YOLO":
        return expected_profit
    else:
        try:
            return float(expected_profit)
        except ValueError:
            click.echo("Not a valid profit input. Enter float or the string, 'YOLO'")
            return None


def valid_diva_chain(chain_id: int) -> bool:
    """Ensure given chain ID supports reporting Diva Protocol data."""
    if chain_id not in DIVA_PROTOCOL_CHAINS:
        print(
            f"Current chain id ({chain_id}) not supported for"
            " reporting Diva Protocol data."
        )
        return False
    return True


def print_reporter_settings(
    signature_address: str,
    query_tag: str,
    gas_limit: int,
    priority_fee: Optional[int],
    expected_profit: str,
    chain_id: int,
    max_fee: Optional[int],
    transaction_type: int,
    legacy_gas_price: Optional[int],
    gas_price_speed: str,
    diva_pool_id: Optional[int],
) -> None:
    """Print user settings to console."""
    click.echo("")

    if signature_address != "":
        click.echo("⚡🤖⚡ Reporting through Flashbots relay ⚡🤖⚡")
        click.echo(f"Signature account: {signature_address}")

    if query_tag:
        click.echo(f"Reporting query tag: {query_tag}")
    elif diva_pool_id is not None:
        click.echo(f"Reporting data for Diva Protocol Pool ID {diva_pool_id}")
    else:
        click.echo("Reporting with synchronized queries")

    click.echo(f"Current chain ID: {chain_id}")

    if expected_profit == "YOLO":
        click.echo("🍜🍜🍜 Reporter not enforcing profit threshold! 🍜🍜🍜")
    else:
        click.echo(f"Expected percent profit: {expected_profit}%")

    click.echo(f"Transaction type: {transaction_type}")
    click.echo(f"Gas Limit: {gas_limit}")
    click.echo(f"Legacy gas price (gwei): {legacy_gas_price}")
    click.echo(f"Max fee (gwei): {max_fee}")
    click.echo(f"Priority fee (gwei): {priority_fee}")
    click.echo(f"Gas price speed: {gas_price_speed}\n")


def reporter_cli_core(ctx: click.Context) -> TelliotCore:
    """Get telliot core configured in reporter CLI context"""
    # Delegate to main cli core getter
    # (handles ACCOUNT_NAME, CHAIN_ID, and TEST_CONFIG)
    core = cli_core(ctx)

    # Ensure chain id compatible with flashbots relay
    if ctx.obj["SIGNATURE_ACCOUNT_NAME"] is not None:
        # Only supports mainnet
        assert core.config.main.chain_id == 1

    assert core.config

    return core


# Main CLI options
@click.group()
@click.option(
    "--account",
    "-a",
    "account",
    help="Name of account used for reporting.",
    required=False,
    nargs=1,
    type=str,
)
@click.option(
    "--signature-account",
    "-sa",
    "signature_account",
    help="Name of signature account used for reporting with Flashbots.",
    required=False,
    nargs=1,
    type=str,
)
@click.option(
    "--test_config",
    is_flag=True,
    help="Runs command with test configuration (developer use only)",
)
@click.pass_context
def cli(
    ctx: Context,
    account: str,
    signature_account: str,
    test_config: bool,
) -> None:
    """Telliot command line interface"""
    ctx.ensure_object(dict)
    ctx.obj["ACCOUNT_NAME"] = account
    ctx.obj["SIGNATURE_ACCOUNT_NAME"] = signature_account
    ctx.obj["TEST_CONFIG"] = test_config

    # Pull chain from account
    # Note: this is not be reliable because accounts can be associated with
    # multiple chains.
    accounts = find_accounts(name=account)
    ctx.obj["CHAIN_ID"] = accounts[0].chains[0]


# Report subcommand options
@cli.command()
@click.option(
    "--query-tag",
    "-qt",
    "query_tag",
    help="select datafeed using query tag",
    required=False,
    nargs=1,
    type=click.Choice([q.tag for q in query_catalog.find()]),
)
@click.option(
    "--gas-limit",
    "-gl",
    "gas_limit",
    help="use custom gas limit",
    nargs=1,
    type=int,
    default=350000,
)
@click.option(
    "--max-fee",
    "-mf",
    "max_fee",
    help="use custom maxFeePerGas (gwei)",
    nargs=1,
    type=int,
    required=False,
)
@click.option(
    "--priority-fee",
    "-pf",
    "priority_fee",
    help="use custom maxPriorityFeePerGas (gwei)",
    nargs=1,
    type=int,
    required=False,
)
@click.option(
    "--gas-price",
    "-gp",
    "legacy_gas_price",
    help="use custom legacy gasPrice (gwei)",
    nargs=1,
    type=int,
    required=False,
)
@click.option(
    "--profit",
    "-p",
    "expected_profit",
    help="lower threshold (inclusive) for expected percent profit",
    nargs=1,
    # User can omit profitability checks by specifying "YOLO"
    type=str,
    default="100.0",
)
@click.option(
    "--tx-type",
    "-tx",
    "tx_type",
    help="choose transaction type (0 for legacy txs, 2 for EIP-1559)",
    type=int,
    default=0,
)
@click.option(
    "--gas-price-speed",
    "-gps",
    "gas_price_speed",
    help="gas price speed for eth gas station API",
    nargs=1,
    type=click.Choice(
        ["safeLow", "average", "fast", "fastest"],
        case_sensitive=True,
    ),
    default="fast",
)
@click.option(
    "--pool-id",
    "-pid",
    "diva_pool_id",
    help="pool ID for Diva Protocol on Polygon",
    nargs=1,
    type=int,
)
@click.option("--submit-once/--submit-continuous", default=False)
@click.option("-pwd", "--password", type=str)
@click.option("-spwd", "--signature-password", type=str)
@click.pass_context
@async_run
async def report(
    ctx: Context,
    query_tag: str,
    tx_type: int,
    gas_limit: int,
    max_fee: Optional[int],
    priority_fee: Optional[int],
    legacy_gas_price: Optional[int],
    expected_profit: str,
    submit_once: bool,
    gas_price_speed: str,
    diva_pool_id: int,
    password: str,
    signature_password: str,
) -> None:
    """Report values to Tellor oracle"""
    # Ensure valid user input for expected profit
    expected_profit = parse_profit_input(expected_profit)  # type: ignore
    if expected_profit is None:
        return

    assert tx_type in (0, 2)

    name = ctx.obj["ACCOUNT_NAME"]
    sig_acct_name = ctx.obj["SIGNATURE_ACCOUNT_NAME"]

    try:
        if not password:
            password = getpass.getpass(f"Enter password for {name} keyfile: ")
    except ValueError:
        click.echo("Invalid Password")

    if sig_acct_name is not None:
        try:
            if not signature_password:
                signature_password = getpass.getpass(
                    f"Enter password for {sig_acct_name} keyfile: "
                )
        except ValueError:
            click.echo("Invalid Password")

    # Initialize telliot core app using CLI context
    async with reporter_cli_core(ctx) as core:

        # Make sure current account is unlocked
        account = core.get_account()
        if not account.is_unlocked:
            account.unlock(password)

        if sig_acct_name is not None:
            sig_account = find_accounts(name=sig_acct_name)[0]
            if not sig_account.is_unlocked:
                sig_account.unlock(password)
            sig_acct_addr = to_checksum_address(sig_account.address)
        else:
            sig_acct_addr = ""  # type: ignore

        cid = core.config.main.chain_id

        # Use selected feed, or choose automatically
        if query_tag is not None:
            chosen_feed = CATALOG_FEEDS[query_tag]
        elif diva_pool_id is not None:
            if not valid_diva_chain(chain_id=cid):
                return
            # Generate datafeed
            chosen_feed = await assemble_diva_datafeed(
                pool_id=diva_pool_id, node=core.endpoint, account=account
            )
        else:
            chosen_feed = None

        print_reporter_settings(
            signature_address=sig_acct_addr,
            query_tag=query_tag,
            transaction_type=tx_type,
            gas_limit=gas_limit,
            max_fee=max_fee,
            priority_fee=priority_fee,
            legacy_gas_price=legacy_gas_price,
            expected_profit=expected_profit,
            chain_id=cid,
            gas_price_speed=gas_price_speed,
            diva_pool_id=diva_pool_id,
        )

        _ = input("Press [ENTER] to confirm settings.")

        common_reporter_kwargs = {
            "endpoint": core.endpoint,
            "account": account,
            "datafeed": chosen_feed,
            "transaction_type": tx_type,
            "gas_limit": gas_limit,
            "max_fee": max_fee,
            "priority_fee": priority_fee,
            "legacy_gas_price": legacy_gas_price,
            "gas_price_speed": gas_price_speed,
            "chain_id": cid,
        }

        # Report to Polygon TellorFlex
        if core.config.main.chain_id in TELLOR_FLEX_CHAINS:
            stake = get_stake_amount()

            tellorflex = core.get_tellorflex_contracts()

            # Type 2 transactions unsupported currently
            common_reporter_kwargs["transaction_type"] = 0

            reporter = PolygonReporter(
                oracle=tellorflex.oracle,
                token=tellorflex.token,
                stake=stake,
                **common_reporter_kwargs,
            )
        # Report to TellorX
        else:
            tellorx = core.get_tellorx_contracts()
            tellorx_reporter_kwargs = {
                "master": tellorx.master,
                "oracle": tellorx.oracle,
                "expected_profit": expected_profit,
                **common_reporter_kwargs,
            }

            if sig_acct_addr != "":
                reporter = FlashbotsReporter(
                    **tellorx_reporter_kwargs,
                    signature_account=sig_account,
                )  # type: ignore
            else:
                reporter = IntervalReporter(**tellorx_reporter_kwargs)  # type: ignore

        if submit_once:
            _, _ = await reporter.report_once()
        else:
            _, _ = await reporter.report()


@cli.command()
@click.option(
    "--query-tag",
    "-qt",
    "query_tag",
    help="select datafeed using query tag",
    required=False,
    nargs=1,
    type=click.Choice([q.tag for q in query_catalog.find()]),
)
@click.option(
    "--amount-trb",
    "-trb",
    "amount_trb",
    help="amount to tip in TRB for a query ID",
    nargs=1,
    type=float,
    required=True,
)
@click.pass_context
@async_run
async def tip(
    ctx: Context,
    query_tag: str,
    amount_trb: float,
) -> None:
    """Tip TRB for a selected query ID"""

    # Initialize telliot core app using CLI context
    async with reporter_cli_core(ctx) as core:

        click.echo(f"Tipping {amount_trb} TRB for query tag: {query_tag}.")

        chosen_feed = CATALOG_FEEDS[query_tag]
        tip = int(amount_trb * 1e18)

        tellorx = core.get_tellorx_contracts()
        tx_receipt, status = asyncio.run(
            tip_query(
                oracle=tellorx.oracle,
                datafeed=chosen_feed,
                tip=tip,
            )
        )

        if status.ok and not status.error and tx_receipt:
            click.echo("Success!")
            tx_hash = tx_receipt["transactionHash"].hex()
            # Point to relevant explorer
            logger.info(f"View tip: \n{core.endpoint.explorer}/tx/{tx_hash}")
        else:
            logger.error(status)


@cli.command()
@click.option(
    "--pool-id",
    "-pid",
    "pool_id",
    help="pool ID for Diva Protocol on Polygon",
    nargs=1,
    type=int,
    required=True,
)
@click.option(
    "--gas-price",
    "-gp",
    "legacy_gas_price",
    help="use custom legacy gasPrice (gwei)",
    nargs=1,
    type=int,
    required=False,
    default=100,
)
@click.option("-pswd", "--password", type=str)
@click.pass_context
@async_run
async def settle(
    ctx: Context,
    pool_id: int,
    password: str,
    legacy_gas_price: int = 100,
) -> None:
    """Settle a derivative pool in DIVA Protocol."""

    name = ctx.obj["ACCOUNT_NAME"]
    try:
        if not password:
            password = getpass.getpass(f"Enter password for {name} keyfile: ")
    except ValueError:
        click.echo("Invalid Password")

    # Initialize telliot core app using CLI context
    async with reporter_cli_core(ctx) as core:

        # Make sure current account is unlocked
        account = core.get_account()
        if not account.is_unlocked:
            account.unlock(password)

        cid = core.config.main.chain_id
        if not valid_diva_chain(chain_id=cid):
            return

        oracle = DivaOracleTellorContract(core.endpoint, account)
        oracle.connect()

        status = await oracle.set_final_reference_value(
            pool_id=pool_id, legacy_gas_price=legacy_gas_price
        )
        if status is not None and status.ok:
            click.echo(f"Pool {pool_id} settled.")
        else:
            click.echo(f"Unable to settle Pool {pool_id}.")


if __name__ == "__main__":
    cli()



from datetime import datetime

import pytest

from telliot_feed_examples.sources.price.historical.cryptowatch import (
    CryptowatchHistoricalPriceService,
)
from telliot_feed_examples.sources.price.historical.kraken import (
    KrakenHistoricalPriceService,
)
from telliot_feed_examples.sources.price.historical.poloniex import (
    PoloniexHistoricalPriceService,
)


def isfloat(num: str) -> bool:
    """Check if string is numeric."""
    try:
        float(num)
        return True
    except ValueError:
        return False


async def get_price(asset, currency, s):
    """Helper function for retrieving prices."""
    v, t = await s.get_price(asset, currency)
    return v, t


def validate_price(v, t):
    """Check types and price anomalies."""
    assert v is not None
    assert isinstance(v, float)
    assert v > 0
    assert isinstance(t, datetime)
    print(v)
    print(t)


@pytest.mark.asyncio
async def test_kraken_get_price():
    """Retrieve singular price close to given timestamp."""
    v, t = await KrakenHistoricalPriceService().get_price("eth", "usd", ts=1616663420)
    validate_price(v, t)

    v, t = await KrakenHistoricalPriceService().get_price("xbt", "usd", ts=1616663420)
    validate_price(v, t)


@pytest.mark.asyncio
async def test_kraken_get_trades():
    """Retrieve all price data given a timestamp and surrounding time period."""
    six_hours = 60 * 60 * 6  # seconds
    trades, t = await KrakenHistoricalPriceService().get_trades(
        "eth",
        "usd",
        period=six_hours,
        ts=1647782323,
    )

    assert isinstance(t, datetime)
    assert isinstance(trades, list)
    assert len(trades) > 0
    assert isfloat(trades[0][0])
    print("# trades in six hour window:", len(trades))

    trades, t = await KrakenHistoricalPriceService().get_trades(
        "xbt",
        "usd",
        period=six_hours,
        ts=1647782323,
    )

    assert isinstance(t, datetime)
    assert isinstance(trades, list)
    assert len(trades) > 0
    assert isfloat(trades[0][0])
    print("# trades in six hour window:", len(trades))


@pytest.mark.asyncio
async def test_poloniex_get_price():
    """Retrieve single historical price close to given timestamp."""
    v, t = await PoloniexHistoricalPriceService().get_price("eth", "dai", ts=1645813159)
    validate_price(v, t)

    v, t = await PoloniexHistoricalPriceService().get_price(
        "eth", "tusd", ts=1645822159
    )

    v, t = await PoloniexHistoricalPriceService().get_price("btc", "dai", ts=1645813159)
    validate_price(v, t)

    v, t = await PoloniexHistoricalPriceService().get_price(
        "btc", "tusd", ts=1645822159
    )


@pytest.mark.asyncio
async def test_poloniex_get_trades():
    """Retrieve all price data given a timestamp and surrounding time period."""
    six_hours = 60 * 60 * 6  # seconds
    trades, t = await PoloniexHistoricalPriceService().get_trades(
        "eth",
        "tusd",
        period=six_hours,
        ts=1647782323,
    )

    assert isinstance(t, datetime)
    assert isinstance(trades, list)
    assert len(trades) > 0
    assert isfloat(trades[0]["rate"])
    print("# trades in six hour window:", len(trades))

    trades, t = await PoloniexHistoricalPriceService().get_trades(
        "btc",
        "tusd",
        period=six_hours,
        ts=1647782323,
    )

    assert isinstance(t, datetime)
    assert isinstance(trades, list)
    assert len(trades) > 0
    assert isfloat(trades[0]["rate"])
    print("# trades in six hour window:", len(trades))


@pytest.mark.skip("TODO: handle candle data for certain timestamps")
@pytest.mark.asyncio
async def test_cryptowatch_get_price():
    """Retrieve single historical price close to given timestamp."""
    v, t = await CryptowatchHistoricalPriceService().get_price(
        "eth", "usd", ts=1648567107
    )
    validate_price(v, t)

    v, t = await CryptowatchHistoricalPriceService().get_price(
        "btc", "usd", ts=1648567107
    )
    validate_price(v, t)


@pytest.mark.skip("TODO: handle candle data for certain timestamps")
@pytest.mark.asyncio
async def test_cryptowatch_get_candles():
    """Retrieve all price data given a timestamp and surrounding time period."""
    six_hours = 60 * 60 * 6  # seconds
    candles, t = await CryptowatchHistoricalPriceService().get_candles(
        "eth",
        "usd",
        period=six_hours,
        ts=1649252172,
    )

    assert isinstance(t, datetime)
    assert isinstance(candles, list)
    assert len(candles) > 0
    assert isfloat(candles[-1][4])
    print("# eth/usd candles in six hour window:", len(candles))

    candles, t = await CryptowatchHistoricalPriceService().get_candles(
        "btc",
        "usd",
        period=six_hours,
        ts=1648567107,
    )

    assert isinstance(t, datetime)
    assert isinstance(candles, list)
    assert len(candles) > 0
    assert isfloat(candles[-1][4])
    print("# btc/usd candles in six hour window:", len(candles))



"""Example datafeed used by USPCEReporter."""
from telliot_core.datafeed import DataFeed
from telliot_core.queries.legacy_query import LegacyRequest

from telliot_feed_examples.sources.uspce import USPCESource


uspce_feed = DataFeed(query=LegacyRequest(legacy_id=41), source=USPCESource())



import asyncio
import csv
from typing import Any

from telliot_feed_examples.sources.price.historical.cryptowatch import (
    CryptowatchHistoricalPriceService,
)
from telliot_feed_examples.sources.price.historical.kraken import (
    KrakenHistoricalPriceService,
)
from telliot_feed_examples.sources.price.historical.poloniex import (
    PoloniexHistoricalPriceService,
)


def print_num_trades(
    service_name: str, asset: str, currency: str, data: list[Any]
) -> None:
    print(
        f"{service_name}: # trades in six hour window for {asset}/{currency}:",
        len(data),
    )


def get_kraken_data(period: int, ts: int, asset: str, currency: str) -> list:
    trades, _ = asyncio.run(
        KrakenHistoricalPriceService().get_trades(
            asset,
            currency,
            period=period,
            ts=ts,
        )
    )
    print_num_trades("Kraken", asset, currency, trades)
    return trades


def get_poloniex_data(period: int, ts: int, asset: str, currency: str) -> list:
    trades, _ = asyncio.run(
        PoloniexHistoricalPriceService().get_trades(asset, currency, period, ts)
    )
    print_num_trades("Poloniex", asset, currency, trades)
    field_idx_lookup = {
        "globalTradeID": 0,
        "tradeID": 1,
        "date": 2,
        "type": 3,
        "rate": 4,
        "amount": 5,
        "total": 6,
        "orderNumber": 7,
    }
    trades_lists = []
    for trade in trades:
        new_trade = [0, 0, 0, 0, 0, 0, 0, 0]
        for key in trade.keys():
            new_trade[field_idx_lookup[key]] = trade[key]
        trades_lists.append(new_trade)
    return trades_lists


def get_cryptowatch_data(period: int, ts: int, asset: str, currency: str) -> list:
    candles, _ = asyncio.run(
        CryptowatchHistoricalPriceService().get_candles(
            asset=asset, currency=currency, period=period, candle_periods=60, ts=ts
        )
    )
    print(
        f"Cryptowatch: # candles in six hour window for {asset}/{currency}:",
        len(candles),
    )
    return candles


def generate_csv(file_name: str, data: list, cols: list) -> None:
    with open(file_name, "w") as f:
        write = csv.writer(f)
        write.writerow(cols)
        write.writerows(data)


def main():
    time_period = 60 * 60 * 6  # Six hours in seconds
    timestamp = 1648567107

    # Kraken historical price data
    # source: https://docs.kraken.com/rest/#operation/getRecentTrades
    cols = ["price", "volume", "time", "buy/sell", "market/limit", "miscellaneous"]

    data = get_kraken_data(
        ts=timestamp, period=time_period, asset="eth", currency="usd"
    )
    generate_csv("kraken_eth_usd_historical_price_source.csv", data, cols)

    data = get_kraken_data(
        ts=timestamp, period=time_period, asset="xbt", currency="usd"
    )
    generate_csv("kraken_xbt_usd_historical_price_source.csv", data, cols)

    # Poloniex historical price data
    # source: https://docs.poloniex.com/#returntradehistory-public
    cols = [
        "globalTradeID",
        "tradeID",
        "date",
        "type",
        "rate",
        "amount",
        "total",
        "orderNumber",
    ]

    data = get_poloniex_data(
        ts=timestamp, period=time_period, asset="eth", currency="tusd"
    )
    generate_csv("poloniex_eth_tusd_historical_price_source.csv", data, cols)

    data = get_poloniex_data(
        ts=timestamp, period=time_period, asset="btc", currency="tusd"
    )
    generate_csv("poloniex_btc_tusd_historical_price_source.csv", data, cols)

    # Cryptowatch historical price data
    # source:
    # cols = ["price", "volume", "time", "buy/sell", "market/limit", "miscellaneous"]
    cols = [
        "CloseTime",
        "OpenPrice",
        "HighPrice",
        "LowPrice",
        "ClosePrice",
        "Volume",
        "QuoteVolume",
    ]
    data = get_cryptowatch_data(
        ts=timestamp, period=time_period, asset="eth", currency="usd"
    )
    generate_csv("cryptowatch_eth_usd_historical_price_source.csv", data, cols)
    data = get_cryptowatch_data(
        ts=timestamp, period=time_period, asset="btc", currency="usd"
    )
    generate_csv("cryptowatch_btc_usd_historical_price_source.csv", data, cols)


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from datetime import datetime
from typing import Any
from typing import Optional
from typing import Tuple
from urllib.parse import urlencode

import requests
from telliot_core.dtypes.datapoint import datetime_now_utc
from telliot_core.dtypes.datapoint import OptionalDataPoint
from telliot_core.pricing.price_source import PriceSource

from telliot_feed_examples.utils.log import get_logger


logger = get_logger(__name__)


# Hardcoded supported asset/currency pairs
# Poloniex swaps the usual order of asset/currency to currency/asset
poloniex_pairs = {"DAI_ETH", "TUSD_ETH", "DAI_BTC", "TUSD_BTC"}


class PoloniexHistoricalPriceService:
    """Poloniex Historical Price Service"""

    def __init__(
        self,
        ts: int = 0,
        timeout: float = 1,
        name: str = "Poloniex Historical Price Service",
        url: str = "https://poloniex.com/",
    ):
        self.name = name
        self.url = url
        self.ts = ts
        self.timeout = timeout

    def get_url(self, url: str = "") -> dict[str, Any]:
        """Helper function to get URL JSON response while handling exceptions

        Args:
            url: URL to fetch

        Returns:
            A dictionary with the following (optional) keys:
                json (dict or list): Result, if no error occurred
                error (str): A description of the error, if one occurred
                exception (Exception): The exception, if one occurred
        """

        request_url = self.url + url

        with requests.Session() as s:
            try:
                r = s.get(request_url, timeout=self.timeout)
                json_data = r.json()
                return {"response": json_data}

            except requests.exceptions.ConnectTimeout as e:
                return {"error": "Timeout Error", "exception": e}

            except Exception as e:
                return {"error": str(type(e)), "exception": e}

    async def get_trades(
        self,
        asset: str,
        currency: str,
        period: int = 900,  # 15 minutes
        ts: Optional[int] = None,
    ) -> Tuple[Optional[list[Any]], Optional[datetime]]:
        """Returns trade data found within a given period.
        The period's middle is at the given timestamp."""
        if ts is None:
            ts = self.ts

        asset = asset.upper()
        currency = currency.upper()
        pair = (
            currency + "_" + asset
        )  # Poloniex wants the reverse of standard order: asset/currency

        if pair not in poloniex_pairs:
            raise Exception(f"Currency pair not supported: {pair}")

        url_params = urlencode(
            {
                "currencyPair": pair,
                "start": int(ts - period),
                "end": ts,
            }
        )

        # Source: https://docs.poloniex.com/#returntradehistory-public
        request_url = f"public?command=returnTradeHistory&{url_params}"

        d = self.get_url(request_url)
        trades = []

        if "error" in d:
            logger.error(d)
            return None, None

        elif "response" in d:
            rsp = d["response"]
            if type(rsp) == dict and "error" in rsp:
                logger.error(f'Error in poloniex response: {rsp["error"]}')
                return None, None
            trades = rsp

        else:
            raise Exception("Invalid response from get_url")

        return trades, datetime_now_utc()

    async def get_price(
        self,
        asset: str,
        currency: str,
        period: int = 20000,
        ts: Optional[int] = None,
    ) -> OptionalDataPoint[float]:
        """Implement PriceServiceInterface

        This implementation gets the historical price from
        the Poloniex API using a timestamp."""
        trades, dt = await self.get_trades(
            asset=asset, currency=currency, ts=ts, period=period
        )

        if trades is not None:
            try:
                if len(trades) == 0:
                    logger.warning(
                        "No data from Poloniex historical price source for"
                        f" given timestamp: {ts}."
                    )
                    return None, None

                # Price from last trade in period
                return float(trades[-1]["rate"]), dt

            except KeyError as e:
                msg = f"Error parsing Poloniex API response: KeyError: {e}"
                logger.critical(msg)

        return None, None


@dataclass
class PoloniexHistoricalPriceSource(PriceSource):
    ts: int = 0
    asset: str = ""
    currency: str = ""
    service: PoloniexHistoricalPriceService = PoloniexHistoricalPriceService(ts=ts)

    def __post_init__(self) -> None:
        self.service.ts = self.ts



from telliot_core.plugin.registry import PluginRegistry

# from telliot_feed_examples.feeds.btc_usd_feed import btc_usd_median_feed

registry = PluginRegistry()
# registry.register_feed(btc_usd_median_feed)

__version__ = "0.0.14"



"""ABIs for contracts."""

gorli_playground_abi = [
    {"inputs": [], "stateMutability": "nonpayable", "type": "constructor"},
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "owner",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "spender",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
        ],
        "name": "Approval",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "bytes32",
                "name": "_queryId",
                "type": "bytes32",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_time",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "bytes",
                "name": "_value",
                "type": "bytes",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_reward",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_nonce",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "bytes",
                "name": "_queryData",
                "type": "bytes",
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "_reporter",
                "type": "address",
            },
        ],
        "name": "NewReport",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "_user",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "bytes32",
                "name": "_queryId",
                "type": "bytes32",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_tip",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_totalTip",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "bytes",
                "name": "_queryData",
                "type": "bytes",
            },
        ],
        "name": "TipAdded",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
        ],
        "name": "Transfer",
        "type": "event",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "name": "addresses",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_owner", "type": "address"},
            {"internalType": "address", "name": "_spender", "type": "address"},
        ],
        "name": "allowance",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_spender", "type": "address"},
            {"internalType": "uint256", "name": "_amount", "type": "uint256"},
        ],
        "name": "approve",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "_account", "type": "address"}],
        "name": "balanceOf",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "uint256", "name": "_timestamp", "type": "uint256"},
        ],
        "name": "beginDispute",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "decimals",
        "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "_user", "type": "address"}],
        "name": "faucet",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "_queryId", "type": "bytes32"}],
        "name": "getCurrentReward",
        "outputs": [
            {"internalType": "uint256", "name": "", "type": "uint256"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "_queryId", "type": "bytes32"}],
        "name": "getNewValueCountbyQueryId",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "uint256", "name": "_index", "type": "uint256"},
        ],
        "name": "getTimestampbyQueryIdandIndex",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "_hash", "type": "bytes32"}],
        "name": "getVoteRounds",
        "outputs": [{"internalType": "uint256[]", "name": "", "type": "uint256[]"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "", "type": "bytes32"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
        ],
        "name": "isDisputed",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "name",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "uint256", "name": "_timestamp", "type": "uint256"},
        ],
        "name": "retrieveData",
        "outputs": [{"internalType": "bytes", "name": "", "type": "bytes"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "bytes", "name": "_value", "type": "bytes"},
            {"internalType": "uint256", "name": "_nonce", "type": "uint256"},
            {"internalType": "bytes", "name": "_queryData", "type": "bytes"},
        ],
        "name": "submitValue",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "symbol",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "timeBasedReward",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "timeOfLastNewValue",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "", "type": "bytes32"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
        ],
        "name": "timestamps",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "uint256", "name": "_amount", "type": "uint256"},
            {"internalType": "bytes", "name": "_queryData", "type": "bytes"},
        ],
        "name": "tipQuery",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "name": "tips",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "tipsInContract",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "totalSupply",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_recipient", "type": "address"},
            {"internalType": "uint256", "name": "_amount", "type": "uint256"},
        ],
        "name": "transfer",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_sender", "type": "address"},
            {"internalType": "address", "name": "_recipient", "type": "address"},
            {"internalType": "uint256", "name": "_amount", "type": "uint256"},
        ],
        "name": "transferFrom",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "", "type": "bytes32"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
        ],
        "name": "values",
        "outputs": [{"internalType": "bytes", "name": "", "type": "bytes"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "voteCount",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "", "type": "bytes32"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
        ],
        "name": "voteRounds",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
]



"""
Tests covering the IntervalReporter class from
telliot's reporters subpackage.
"""
from datetime import datetime
from typing import Any

import pytest
from telliot_core.apps.core import TelliotCore
from telliot_core.datafeed import DataFeed
from telliot_core.gas.etherscan_gas import EtherscanGasPrice
from telliot_core.utils.response import ResponseStatus
from web3.datastructures import AttributeDict

from telliot_feed_examples.feeds.eth_usd_feed import eth_usd_median_feed
from telliot_feed_examples.reporters.interval import IntervalReporter


@pytest.fixture
async def eth_usd_reporter(rinkeby_cfg):
    """Returns an instance of an IntervalReporter using
    the ETH/USD median datafeed."""
    async with TelliotCore(config=rinkeby_cfg) as core:
        account = core.get_account()
        tellorx = core.get_tellorx_contracts()
        r = IntervalReporter(
            endpoint=core.config.get_endpoint(),
            account=account,
            master=tellorx.master,
            oracle=tellorx.oracle,
            datafeed=eth_usd_median_feed,
            expected_profit="YOLO",
            transaction_type=0,
            gas_limit=400000,
            max_fee=None,
            priority_fee=None,
            legacy_gas_price=None,
            gas_price_speed="safeLow",
            chain_id=core.config.main.chain_id,
        )
        return r


@pytest.mark.asyncio
async def test_fetch_datafeed(eth_usd_reporter):
    r = eth_usd_reporter
    feed = await r.fetch_datafeed()
    assert isinstance(feed, DataFeed)

    r.datafeed = None
    assert r.datafeed is None
    feed = await r.fetch_datafeed()
    assert isinstance(feed, DataFeed)


@pytest.mark.skip(
    "Skipping because the error is from telliot-core and not from the reporter."
)
@pytest.mark.asyncio
async def test_get_fee_info(eth_usd_reporter):
    info, time = await eth_usd_reporter.get_fee_info()

    assert isinstance(time, datetime)
    assert isinstance(info, EtherscanGasPrice)
    assert isinstance(info.LastBlock, int)
    assert info.LastBlock > 0
    assert isinstance(info.gasUsedRatio, list)


@pytest.mark.asyncio
async def test_get_num_reports_by_id(eth_usd_reporter):
    r = eth_usd_reporter
    num, status = await r.get_num_reports_by_id(r.datafeed.query.query_id)

    assert isinstance(status, ResponseStatus)

    if status.ok:
        assert isinstance(num, int)
    else:
        assert num is None


@pytest.mark.asyncio
async def test_ensure_staked(eth_usd_reporter):
    """Test staking status of reporter."""
    staked, status = await eth_usd_reporter.ensure_staked()

    assert staked
    assert status.ok


@pytest.mark.asyncio
async def test_check_reporter_lock(eth_usd_reporter):
    """Test checking if in reporter lock."""
    r = eth_usd_reporter
    r.last_submission_timestamp = 0

    status = await r.check_reporter_lock()

    assert isinstance(status, ResponseStatus)
    if not status.ok:
        assert status.error == "Current address is in reporter lock."
        assert r.last_submission_timestamp != 0


@pytest.mark.asyncio
async def test_ensure_profitable(eth_usd_reporter):
    """Test profitability check."""
    r = eth_usd_reporter

    assert r.expected_profit == "YOLO"

    status = await r.ensure_profitable(r.datafeed)

    assert status.ok

    r.expected_profit = 1e10
    status = await r.ensure_profitable(r.datafeed)

    assert not status.ok
    assert status.error == "Estimated profitability below threshold."


@pytest.mark.asyncio
async def test_fetch_gas_price(eth_usd_reporter):
    """Test retrieving custom gas price from eth gas station."""
    r = eth_usd_reporter

    assert r.gas_price_speed == "safeLow"

    r.gas_price_speed = "average"

    assert r.gas_price_speed != "safeLow"

    gas_price = await r.fetch_gas_price()

    assert isinstance(gas_price, int)
    assert gas_price > 0


@pytest.mark.skip("Asks for psswd")
@pytest.mark.asyncio
async def test_interval_reporter_submit_once(eth_usd_reporter):
    """Test reporting once to the TellorX playground on Rinkeby
    with three retries."""
    r = eth_usd_reporter

    # Sync reporter
    r.datafeed = None

    EXPECTED_ERRORS = {
        "Current addess disputed. Switch address to continue reporting.",
        "Current address is locked in dispute or for withdrawal.",
        "Current address is in reporter lock.",
        "Estimated profitability below threshold.",
        "Estimated gas price is above maximum gas price.",
        "Unable to retrieve updated datafeed value.",
    }

    ORACLE_ADDRESSES = {
        "0xe8218cACb0a5421BC6409e498d9f8CC8869945ea",  # mainnet
        "0x18431fd88adF138e8b979A7246eb58EA7126ea16",  # rinkeby
    }

    tx_receipt, status = await r.report_once()

    # Reporter submitted
    if tx_receipt is not None and status.ok:
        assert isinstance(tx_receipt, AttributeDict)
        assert tx_receipt.to in ORACLE_ADDRESSES
    # Reporter did not submit
    else:
        assert not tx_receipt
        assert not status.ok
        assert status.error in EXPECTED_ERRORS


@pytest.mark.skip("Asks for psswd")
@pytest.mark.asyncio
async def test_no_updated_value(eth_usd_reporter, bad_source):
    """Test handling for no updated value returned from datasource."""
    r = eth_usd_reporter

    # Clear latest datapoint
    r.datafeed.source._history.clear()

    # Replace PriceAggregator's sources with test source that
    # returns no updated DataPoint
    r.datafeed.source.sources = [bad_source]

    # Override reporter lock status
    async def passing():
        return ResponseStatus()

    r.check_reporter_lock = passing

    # Override reporter profit check
    async def profit(datafeed: DataFeed[Any]):
        return ResponseStatus()

    r.ensure_profitable = profit

    tx_receipt, status = await r.report_once()

    assert not tx_receipt
    assert not status.ok
    assert status.error == "Unable to retrieve updated datafeed value."



from datetime import datetime

import pytest
from telliot_core.api import DataFeed
from telliot_core.apps.core import TelliotCore
from telliot_core.queries.diva_protocol import divaProtocolPolygon

from telliot_feed_examples.feeds.diva_protocol_feed import assemble_diva_datafeed
from telliot_feed_examples.feeds.diva_protocol_feed import DivaPoolParameters
from telliot_feed_examples.feeds.diva_protocol_feed import get_pool_params
from telliot_feed_examples.feeds.diva_protocol_feed import get_source
from telliot_feed_examples.sources.price.historical.poloniex import (
    PoloniexHistoricalPriceSource,
)


def test_get_source() -> None:
    source = get_source("btc", 1243)

    assert source.sources[1].asset == "xbt"
    assert source.sources[3].ts == 1243


@pytest.mark.asyncio
async def test_get_pool_parameters(ropsten_cfg) -> None:
    async with TelliotCore(config=ropsten_cfg) as core:
        account = core.get_account()
        params = await get_pool_params(3, core.endpoint, account)

        assert isinstance(params, DivaPoolParameters)
        assert params.reference_asset == "ETH/USD"
        assert params.expiry_date == 1657349074


@pytest.mark.asyncio
async def test_diva_datafeed(ropsten_cfg) -> None:
    async with TelliotCore(config=ropsten_cfg) as core:
        account = core.get_account()
        feed = await assemble_diva_datafeed(
            pool_id=3, node=core.endpoint, account=account
        )

        assert isinstance(feed, DataFeed)
        assert isinstance(feed.query, divaProtocolPolygon)
        assert isinstance(feed.source.sources[3], PoloniexHistoricalPriceSource)
        assert isinstance(feed.source.sources[0].ts, int)
        assert feed.source.asset == "eth"
        assert feed.source.sources[2].currency == "dai"

        v, t = await feed.source.fetch_new_datapoint()
        assert v is None or isinstance(v, float)
        if v is not None:
            assert v > 1000
            assert isinstance(t, datetime)

        feed = await assemble_diva_datafeed(
            pool_id=10, node=core.endpoint, account=account
        )

        assert feed.source.asset == "btc"
        assert feed.source.sources[1].asset == "xbt"

        v, t = await feed.source.fetch_new_datapoint()

        assert v is None or isinstance(v, float)
        if v is not None:
            assert v > 30000
            assert isinstance(t, datetime)



from telliot_core.datafeed import DataFeed
from telliot_core.queries import SpotPrice

from telliot_feed_examples.sources.price.spot.coingecko import CoinGeckoSpotPriceSource
from telliot_feed_examples.sources.price.spot.coinmarketcap import (
    CoinMarketCapSpotPriceSource,
)
from telliot_feed_examples.sources.price.spot.nomics import NomicsSpotPriceSource
from telliot_feed_examples.sources.price_aggregator import PriceAggregator


sources = [
    CoinGeckoSpotPriceSource(asset="bct", currency="usd"),
    NomicsSpotPriceSource(asset="bct", currency="usd"),
    CoinMarketCapSpotPriceSource(asset="bct", currency="usd"),
]


bct_usd_median_feed = DataFeed(
    query=SpotPrice(asset="BCT", currency="USD"),
    source=PriceAggregator(
        asset="bct",
        currency="usd",
        algorithm="median",
        sources=sources,
    ),
)



# Majority of this code from web3-flashbots:
# https://github.com/flashbots/web3-flashbots
# EIP-1559 subbport by @lekhovitsky
# https://github.com/lekhovitsky
# type: ignore
import time
from functools import reduce
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import rlp
from eth_account import Account
from eth_account._utils.legacy_transactions import encode_transaction
from eth_account._utils.legacy_transactions import (
    serializable_unsigned_transaction_from_dict,
)
from eth_account._utils.legacy_transactions import Transaction
from eth_account._utils.typed_transactions import AccessListTransaction
from eth_account._utils.typed_transactions import DynamicFeeTransaction
from eth_typing import HexStr
from hexbytes import HexBytes
from toolz import dissoc
from web3 import Web3
from web3.method import Method
from web3.module import Module
from web3.types import _Hash32
from web3.types import Nonce
from web3.types import RPCEndpoint
from web3.types import TxParams

from .types import FlashbotsBundleDictTx
from .types import FlashbotsBundleRawTx
from .types import FlashbotsBundleTx
from .types import FlashbotsOpts


SECONDS_PER_BLOCK = 15


class FlashbotsRPC:
    eth_sendBundle = RPCEndpoint("eth_sendBundle")
    eth_callBundle = RPCEndpoint("eth_callBundle")


class FlashbotsTransactionResponse:
    w3: Web3
    bundle: List[Any]
    target_block_number: int

    def __init__(self, w3: Web3, txs: List[HexBytes], target_block_number: int):
        self.w3 = w3

        def parse_tx(tx):
            return {
                "signed_transaction": tx,
                "hash": self.w3.keccak(tx),
            }

        self.bundle = list(map(parse_tx, txs))
        self.target_block_number = target_block_number

    def wait(self) -> None:
        """Waits until the target block has been reached"""
        while self.w3.eth.blockNumber < self.target_block_number:
            time.sleep(1)

    def receipts(self) -> List[Union[_Hash32, HexBytes, HexStr]]:
        """Returns all the transaction receipts from the submitted bundle"""
        self.wait()
        return list(
            map(lambda tx: self.w3.eth.getTransactionReceipt(tx["hash"]), self.bundle)
        )


class Flashbots(Module):
    signed_txs: List[HexBytes]
    response: FlashbotsTransactionResponse

    def sign_bundle(
        self,
        bundled_transactions: List[
            Union[FlashbotsBundleTx, FlashbotsBundleRawTx, FlashbotsBundleDictTx]
        ],
    ) -> List[HexBytes]:
        """Given a bundle of signed and unsigned transactions, it signs them all"""
        nonces: Dict[HexStr, Nonce] = {}
        signed_transactions: List[HexBytes] = []

        for tx in bundled_transactions:
            if "signed_transaction" in tx:  # FlashbotsBundleRawTx
                tx_params = _parse_signed_tx(tx["signed_transaction"])
                nonces[tx_params["from"]] = tx_params["nonce"] + 1
                signed_transactions.append(tx["signed_transaction"])

            elif "signer" in tx:  # FlashbotsBundleTx
                signer, tx = tx["signer"], tx["transaction"]
                tx["from"] = signer.address

                if tx.get("nonce") is None:
                    tx["nonce"] = nonces.get(
                        signer.address,
                        self.web3.eth.get_transaction_count(signer.address),
                    )
                nonces[signer.address] = tx["nonce"] + 1

                if "gas" not in tx:
                    tx["gas"] = self.web3.eth.estimateGas(tx)

                signed_tx = signer.sign_transaction(tx)
                signed_transactions.append(signed_tx.rawTransaction)

            elif all(key in tx for key in ["v", "r", "s"]):  # FlashbotsBundleDictTx
                v, r, s = (
                    tx["v"],
                    int(tx["r"].hex(), base=16),
                    int(tx["s"].hex(), base=16),
                )

                tx_dict = {
                    "nonce": tx["nonce"],
                    "to": HexBytes(tx["to"]) if "to" in tx else None,
                    "data": HexBytes(tx["input"]),
                    "value": tx["value"],
                    "gas": tx["gas"],
                }

                if "maxFeePerGas" in tx and "maxPriorityFeePerGas" in tx:
                    tx_dict["maxFeePerGas"], tx_dict["maxPriorityFeePerGas"] = (
                        tx["maxFeePerGas"],
                        tx["maxPriorityFeePerGas"],
                    )
                else:
                    tx_dict["gasPrice"] = tx["gasPrice"]

                if tx.get("accessList"):
                    tx_dict["accessList"] = tx["accessList"]

                unsigned_tx = serializable_unsigned_transaction_from_dict(tx_dict)
                raw = encode_transaction(unsigned_tx, vrs=(v, r, s))
                assert self.web3.keccak(raw) == tx["hash"]
                signed_transactions.append(raw)

        return signed_transactions

    def to_hex(self, signed_transaction: bytes) -> str:
        tx_hex = signed_transaction.hex()
        if tx_hex[0:2] != "0x":
            tx_hex = f"0x{tx_hex}"
        return tx_hex

    def send_raw_bundle_munger(
        self,
        signed_bundled_transactions: List[HexBytes],
        target_block_number: int,
        opts: Optional[FlashbotsOpts] = None,
    ) -> List[Any]:
        """Given a raw signed bundle,
        it packages it up with the block numbre and the timestamps"""
        # convert to hex
        return [
            {
                "txs": list(map(lambda x: self.to_hex(x), signed_bundled_transactions)),
                "blockNumber": hex(target_block_number),
                "minTimestamp": opts["minTimestamp"] if opts else 0,
                "maxTimestamp": opts["maxTimestamp"] if opts else 0,
                "revertingTxHashes": opts["revertingTxHashes"] if opts else [],
            }
        ]

    sendRawBundle: Method[Callable[[Any], Any]] = Method(
        FlashbotsRPC.eth_sendBundle, mungers=[send_raw_bundle_munger]
    )
    send_raw_bundle = sendRawBundle

    def send_bundle_munger(
        self,
        bundled_transactions: List[Union[FlashbotsBundleTx, FlashbotsBundleRawTx]],
        target_block_number: int,
        opts: Optional[FlashbotsOpts] = None,
    ) -> List[Any]:
        signed_txs = self.sign_bundle(bundled_transactions)
        self.response = FlashbotsTransactionResponse(
            self.web3, signed_txs, target_block_number
        )
        return self.send_raw_bundle_munger(signed_txs, target_block_number, opts)

    def raw_bundle_formatter(self, resp) -> Any:
        return lambda _: resp.response

    sendBundle: Method[Callable[[Any], Any]] = Method(
        FlashbotsRPC.eth_sendBundle,
        mungers=[send_bundle_munger],
        result_formatters=raw_bundle_formatter,
    )
    send_bundle = sendBundle

    def simulate(
        self,
        bundled_transactions,
        block_tag: int = None,
        state_block_tag: int = None,
        block_timestamp: int = None,
    ):
        # get block details
        block_details = (
            self.web3.eth.get_block(block_tag)
            if block_tag is not None
            else self.web3.eth.get_block("latest")
        )

        # sets evm params
        evm_block_number = self.web3.toHex(block_details.number)
        evm_block_state_number = (
            state_block_tag
            if state_block_tag is not None
            else self.web3.toHex(block_details.number - 1)
        )
        evm_timestamp = (
            block_timestamp
            if block_timestamp is not None
            else self.extrapolate_timestamp(block_tag, block_details.number)
        )

        signed_bundled_transactions = self.sign_bundle(bundled_transactions)
        # calls evm simulator
        call_result = self.call_bundle(
            signed_bundled_transactions,
            evm_block_number,
            evm_block_state_number,
            evm_timestamp,
        )

        return {
            "bundleHash": call_result["bundleHash"],
            "coinbaseDiff": call_result["coinbaseDiff"],
            "results": call_result["results"],
            "signedBundledTransactions": signed_bundled_transactions,
            "totalGasUsed": reduce(
                lambda a, b: a + b["gasUsed"], call_result["results"], 0
            ),
        }

    def extrapolate_timestamp(self, block_tag: int, latest_block_number: int):
        block_delta = block_tag - latest_block_number
        if block_delta < 0:
            raise Exception("block extrapolation negative")
        return self.web3.eth.get_block(latest_block_number)["timestamp"] + (
            block_delta * SECONDS_PER_BLOCK
        )

    def call_bundle_munger(
        self,
        signed_bundled_transactions: List[
            Union[FlashbotsBundleTx, FlashbotsBundleRawTx]
        ],
        evm_block_number,
        evm_block_state_number,
        evm_timestamp,
        opts: Optional[FlashbotsOpts] = None,
    ) -> Any:
        """Given a raw signed bundle,
        it packages it up with the block number and the timestamps"""
        inpt = [
            {
                "txs": list(map(lambda x: x.hex(), signed_bundled_transactions)),
                "blockNumber": evm_block_number,
                "stateBlockNumber": evm_block_state_number,
                "timestamp": evm_timestamp,
            }
        ]
        return inpt

    call_bundle: Method[Callable[[Any], Any]] = Method(
        json_rpc_method=FlashbotsRPC.eth_callBundle, mungers=[call_bundle_munger]
    )


def _parse_signed_tx(signed_tx: HexBytes) -> TxParams:
    # decode tx params based on its type
    tx_type = signed_tx[0]
    if tx_type > int("0x7f", 16):
        # legacy and EIP-155 transactions
        decoded_tx = rlp.decode(signed_tx, Transaction).as_dict()
    else:
        # typed transactions (EIP-2718)
        if tx_type == 1:
            # EIP-2930
            sedes = AccessListTransaction._signed_transaction_serializer
        elif tx_type == 2:
            # EIP-1559
            sedes = DynamicFeeTransaction._signed_transaction_serializer
        else:
            raise ValueError(f"Unknown transaction type: {tx_type}.")
        decoded_tx = rlp.decode(signed_tx[1:], sedes).as_dict()

    # recover sender address and remove signature fields
    decoded_tx["from"] = Account.recover_transaction(signed_tx)
    decoded_tx = dissoc(decoded_tx, "v", "r", "s")
    return decoded_tx



from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Optional

from pydantic import BaseModel
from telliot_core.dtypes.datapoint import datetime_now_utc
from telliot_core.dtypes.datapoint import OptionalDataPoint
from telliot_core.pricing.price_service import WebPriceService
from telliot_core.pricing.price_source import PriceSource

from telliot_feed_examples.utils.log import get_logger


logger = get_logger(__name__)


class BittrexQuote(BaseModel):
    Bid: float
    Ask: float
    Last: float


class PriceResponse(BaseModel):
    success: bool
    message: str
    result: Optional[BittrexQuote]


class BittrexSpotPriceService(WebPriceService):
    """Bittrex Price Service"""

    def __init__(self, **kwargs: Any):
        super().__init__(
            name="Bittrex Price Service", url="https://api.bittrex.com", **kwargs
        )

    async def get_price(self, asset: str, currency: str) -> OptionalDataPoint[float]:
        """Implement PriceServiceInterface

        This implementation gets the price from the Bittrex API

        Note that the timestamp returned form the coinbase API could be used
        instead of the locally generated timestamp.
        """

        request_url = "/api/v1.1/public/getticker?market={}-{}".format(
            currency.lower(), asset.lower()
        )

        d = self.get_url(request_url)

        if "error" in d:
            logger.error(d)
            return None, None

        else:
            r = PriceResponse.parse_obj(d["response"])
            if r.success:
                if r.result is not None:
                    return r.result.Last, datetime_now_utc()
                else:
                    return None, None
            else:
                logger.error(r.message)
                return None, None


@dataclass
class BittrexSpotPriceSource(PriceSource):
    asset: str = ""
    currency: str = ""
    service: BittrexSpotPriceService = field(
        default_factory=BittrexSpotPriceService, init=False
    )



import json
from dataclasses import dataclass
from dataclasses import field
from typing import Any

from requests import Session
from requests.exceptions import ConnectionError
from requests.exceptions import Timeout
from requests.exceptions import TooManyRedirects
from telliot_core.apps.telliot_config import TelliotConfig
from telliot_core.dtypes.datapoint import datetime_now_utc
from telliot_core.dtypes.datapoint import OptionalDataPoint
from telliot_core.pricing.price_service import WebPriceService
from telliot_core.pricing.price_source import PriceSource

from telliot_feed_examples.utils.log import get_logger

logger = get_logger(__name__)

coinmarketcap_assets = {"BCT"}
coinmarketcap_currencies = {"USD"}

API_KEY = TelliotConfig().api_keys.find(name="coinmarketcap")[0].key


class CoinMarketCapSpotPriceService(WebPriceService):
    """CoinMarketCap Price Service"""

    def __init__(self, **kwargs: Any) -> None:
        kwargs["name"] = "CoinMarketCap Price Service"
        kwargs[
            "url"
        ] = "https://pro-api.coinmarketcap.com/v1/cryptocurrency/quotes/latest"
        super().__init__(**kwargs)

    async def get_price(self, asset: str, currency: str) -> OptionalDataPoint[float]:

        # check for api key in config
        if API_KEY == "":
            logger.warn(
                "To use the CoinMarketCap source, add CoinMarketCap api key to ampl.yaml"
            )
            return None, None

        asset = asset.upper()
        currency = currency.upper()

        if asset not in coinmarketcap_assets:
            raise Exception(f"Asset not supported: {asset}")
        if currency not in coinmarketcap_currencies:
            raise Exception(f"Currency not supported: {currency}")

        request_url = (
            "https://pro-api.coinmarketcap.com/v1/cryptocurrency/quotes/latest"
        )

        parameters = {"symbol": asset}
        headers = {
            "Accepts": "application/json",
            "X-CMC_PRO_API_KEY": API_KEY,
        }

        session = Session()
        session.headers.update(headers)

        try:
            response = session.get(request_url, params=parameters)
            data = json.loads(response.text)

        except (ConnectionError, Timeout, TooManyRedirects) as e:
            logger.warning(e)
            return None, None

        price = data["data"][asset]["quote"][currency]["price"]
        return price, datetime_now_utc()


@dataclass
class CoinMarketCapSpotPriceSource(PriceSource):
    asset: str = ""
    currency: str = ""
    service: CoinMarketCapSpotPriceService = field(
        default_factory=CoinMarketCapSpotPriceService, init=False
    )



from telliot_core.datafeed import DataFeed
from telliot_core.queries import SpotPrice

from telliot_feed_examples.sources.price.spot.coingecko import CoinGeckoSpotPriceSource
from telliot_feed_examples.sources.price.spot.nomics import NomicsSpotPriceSource
from telliot_feed_examples.sources.price_aggregator import PriceAggregator

idle_usd_median_feed = DataFeed(
    query=SpotPrice(asset="IDLE", currency="USD"),
    source=PriceAggregator(
        asset="idle",
        currency="usd",
        algorithm="median",
        sources=[
            CoinGeckoSpotPriceSource(asset="idle", currency="usd"),
            NomicsSpotPriceSource(asset="idle", currency="usd"),
        ],
    ),
)



import statistics

import pytest
from telliot_core.queries.query import OracleQuery

from telliot_feed_examples.feeds.btc_usd_feed import btc_usd_median_feed


@pytest.mark.asyncio
async def test_AssetPriceFeed():
    """Retrieve median BTC price from example datafeed &
    make sure value is within tolerance."""

    # Get query
    q = btc_usd_median_feed.query
    assert isinstance(q, OracleQuery)

    # Fetch price
    # status, price, tstamp = await btc_usd_median_feed.update_value()
    v, t = await btc_usd_median_feed.source.fetch_new_datapoint()

    # Make sure error is less than decimal tolerance
    # assert status.ok
    assert 10000 < v < 100000
    print(f"BTC Price: {v}")

    # Get list of data sources from sources dict
    source_prices = [source.latest[0] for source in btc_usd_median_feed.source.sources]

    # Make sure error is less than decimal tolerance
    assert (v - statistics.median(source_prices)) < 10**-6



from dataclasses import dataclass
from dataclasses import field
from typing import Any
from urllib.parse import urlencode

from telliot_core.apps.telliot_config import TelliotConfig
from telliot_core.dtypes.datapoint import datetime_now_utc
from telliot_core.dtypes.datapoint import OptionalDataPoint
from telliot_core.pricing.price_service import WebPriceService
from telliot_core.pricing.price_source import PriceSource

from telliot_feed_examples.utils.log import get_logger


logger = get_logger(__name__)

# Coinbase API uses the 'id' field from /coins/list.
# Using a manual mapping for now.
nomics_coin_id = {"bct": "BCT5", "btc": "BTC", "ric": "RIC4", "idle": "IDLE"}

API_KEY = TelliotConfig().api_keys.find(name="nomics")[0].key


class NomicsSpotPriceService(WebPriceService):
    """Nomics Price Service"""

    def __init__(self, **kwargs: Any) -> None:
        kwargs["name"] = "Nomics Price Service"
        kwargs["url"] = "https://api.nomics.com"
        super().__init__(**kwargs)

    async def get_price(self, asset: str, currency: str) -> OptionalDataPoint[float]:
        """Implement PriceServiceInterface

        This implementation gets the price from the Nomics API

        """

        if API_KEY == "":
            logger.warning(
                "To use the nomics source, add nomics api key to api_keys.yaml"
            )
            return None, None

        asset = asset.lower()
        currency = currency.upper()

        coin_id = nomics_coin_id.get(asset, None)
        if not coin_id:
            raise Exception("Asset not supported: {}".format(asset))

        url_params = urlencode(
            {
                "key": API_KEY,
                "ids": coin_id.upper(),
                "interval": "1d",
                "convert": currency,
                "per-page": 100,
            }
        )
        request_url = "/v1/currencies/ticker?{}".format(url_params)
        d = self.get_url(request_url)

        if "error" in d:
            logger.error(d)
            return None, None

        elif "response" in d:
            response = d["response"]

            try:
                price = float(response[0]["price"])
            except KeyError as e:
                msg = "Error parsing Nomics API response: KeyError: {}".format(e)
                logger.critical(msg)

        else:
            raise Exception("Invalid response from get_url")

        return price, datetime_now_utc()


@dataclass
class NomicsSpotPriceSource(PriceSource):
    asset: str = ""
    currency: str = ""
    service: NomicsSpotPriceService = field(
        default_factory=NomicsSpotPriceService, init=False
    )



from dataclasses import dataclass
from datetime import datetime
from datetime import timezone
from typing import Any

from telliot_core.datasource import DataSource
from telliot_core.dtypes.datapoint import DataPoint

from telliot_feed_examples.utils.log import get_logger


logger = get_logger(__name__)


@dataclass
class DivaManualSource(DataSource[Any]):
    """DataSource for Diva Protocol manually-entered data."""

    reference_asset: str = ""
    timestamp: int = 0

    def parse_user_val(self) -> float:
        """Parse historical price from user input."""
        print(
            "Enter price to report for reference asset "
            f"{self.reference_asset} at timestamp {self.timestamp}:"
        )

        data = None
        while data is None:
            inpt = input()

            try:
                inpt = float(inpt)  # type: ignore
            except ValueError:
                print("Invalid input. Enter decimal value (float).")
                continue

            print(f"Submitting value: {inpt}\nPress [ENTER] to confirm.")
            _ = input()
            data = inpt

        return data

    async def fetch_new_datapoint(self) -> DataPoint[float]:
        """Update current value with time-stamped value fetched from user input.

        Returns:
            Current time-stamped value
        """
        data = self.parse_user_val()
        dt = datetime.fromtimestamp(self.timestamp, tz=timezone.utc)
        datapoint = (data, dt)

        self.store_datapoint(datapoint)

        logger.info(f"Stored price of {self.reference_asset} at {dt}: {data}")

        return datapoint



"""Datafeed for current price of TRB in USD used by LegacyQueryReporter."""
from telliot_core.datafeed import DataFeed
from telliot_core.queries import LegacyRequest

from telliot_feed_examples.sources.price.spot.coingecko import CoinGeckoSpotPriceSource
from telliot_feed_examples.sources.price_aggregator import PriceAggregator

trb_usd_median_feed = DataFeed(
    query=LegacyRequest(legacy_id=50),
    source=PriceAggregator(
        asset="trb",
        currency="usd",
        algorithm="median",
        sources=[
            CoinGeckoSpotPriceSource(asset="trb", currency="usd"),
        ],
    ),
)



from dataclasses import dataclass
from datetime import datetime
from typing import Any
from typing import Optional
from typing import Tuple
from urllib.parse import urlencode

import requests
from telliot_core.dtypes.datapoint import datetime_now_utc
from telliot_core.dtypes.datapoint import OptionalDataPoint
from telliot_core.pricing.price_service import WebPriceService
from telliot_core.pricing.price_source import PriceSource

from telliot_feed_examples.utils.log import get_logger


logger = get_logger(__name__)


# Hardcoded supported asset/currency pairs
CRYPTOWATCH_PAIRS = {"ethusd", "btcusd"}


class CryptowatchHistoricalPriceService(WebPriceService):
    """Cryptowatch Historical Price Service"""

    def __init__(
        self,
        timeout: float = 0.5,
        name: str = "Cryptowatch Historical Price Service",
        url: str = "https://api.cryptowat.ch/",
        ts: int = 0,
    ):
        self.name = name
        self.url = url
        self.ts = ts
        self.timeout = timeout

    def get_url(self, url: str = "") -> dict[str, Any]:
        """Helper function to get URL JSON response while handling exceptions

        Args:
            url: URL to fetch

        Returns:
            A dictionary with the following (optional) keys:
                json (dict or list): Result, if no error occurred
                error (str): A description of the error, if one occurred
                exception (Exception): The exception, if one occurred
        """

        request_url = self.url + url

        with requests.Session() as s:
            try:
                r = s.get(request_url, timeout=self.timeout)
                json_data = r.json()
                return {"response": json_data}

            except requests.exceptions.ConnectTimeout as e:
                return {"error": "Timeout Error", "exception": e}

            except Exception as e:
                return {"error": str(type(e)), "exception": e}

    async def get_candles(
        self,
        asset: str,
        currency: str,
        candle_periods: int = 60,  # 1 minute
        period: int = 900,  # 15 minutes
        ts: Optional[int] = None,
    ) -> Tuple[Optional[list[Any]], Optional[datetime]]:
        """Implement PriceServiceInterface

        This implementation gets the historical price from
        the Cryptowatch API using a timestamp. Historical prices are
        fetched from Cryptowatch's recorded Coinbase-pro data.

        Documentation for Cryptowatch API:
        https://docs.cryptowat.ch/rest-api/markets/ohlc
        """
        if ts is None:
            ts = self.ts

        asset = asset.lower()
        currency = currency.lower()
        pair = asset + currency

        if pair not in CRYPTOWATCH_PAIRS:
            raise Exception(f"Currency pair not supported: {pair}")

        url_params = urlencode(
            {
                "after": int(ts - period),
                "before": ts,
                "periods": candle_periods,
            }
        )

        request_url = f"markets/coinbase-pro/{pair}/ohlc?{url_params}"

        d = self.get_url(request_url)
        candles = None

        if "error" in d:
            logger.error(d)
            return None, None

        elif "response" in d:
            response = d["response"]

            try:
                candles = response["result"][str(candle_periods)]
            except KeyError as e:
                msg = f"Error parsing Cryptowatch API response: KeyError: {e}"
                logger.critical(msg)

        else:
            raise Exception("Invalid response from get_url")

        return candles, datetime_now_utc()

    async def get_price(
        self, asset: str, currency: str, period: int = 10000, ts: Optional[int] = None
    ) -> OptionalDataPoint[float]:
        """Implement PriceServiceInterface

        This implementation gets the historical price from
        the Cryptowatch API using a timestamp. Historical prices are
        fetched from Cryptowatch's recorded Coinbase-pro data.

        Documentation for Cryptowatch API:
        https://docs.cryptowat.ch/rest-api/markets/ohlc
        """
        candles, dt = await self.get_candles(
            asset=asset, currency=currency, ts=ts, period=period
        )

        if candles is not None:
            try:
                if len(candles) == 0:
                    logger.warning(
                        "No candle data from Cryptowatch historical price source"
                        f" for given timestamp: {ts}."
                    )
                    return None, None

                # Price from last candle in period
                return float(candles[-1][4]), dt

            except KeyError as e:
                msg = f"Error parsing Cryptowatch API candle data: KeyError: {e}"
                logger.critical(msg)

        return None, None


@dataclass
class CryptowatchHistoricalPriceSource(PriceSource):
    ts: int = 0
    asset: str = ""
    currency: str = ""
    service: CryptowatchHistoricalPriceService = CryptowatchHistoricalPriceService(
        ts=ts
    )

    def __post_init__(self) -> None:
        self.service.ts = self.ts



"""Helper functions for reporting data for Diva Protocol."""
import logging
from dataclasses import dataclass
from typing import Optional

from chained_accounts import ChainedAccount
from telliot_core.api import DataFeed
from telliot_core.model.endpoints import RPCEndpoint
from telliot_core.queries.diva_protocol import divaProtocolPolygon
from telliot_core.tellor.tellorflex.diva import DivaProtocolContract

from telliot_feed_examples.sources.price.historical.cryptowatch import (
    CryptowatchHistoricalPriceSource,
)
from telliot_feed_examples.sources.price.historical.kraken import (
    KrakenHistoricalPriceSource,
)
from telliot_feed_examples.sources.price.historical.poloniex import (
    PoloniexHistoricalPriceSource,
)
from telliot_feed_examples.sources.price_aggregator import PriceAggregator


logger = logging.getLogger(__name__)


SUPPORTED_REFERENCE_ASSETS = {"ETH/USD", "BTC/USD"}


@dataclass
class DivaPoolParameters:
    """More info: https://github.com/divaprotocol/oracles#diva-smart-contract"""

    reference_asset: str
    expiry_date: int


async def get_pool_params(
    pool_id: int, node: RPCEndpoint, account: ChainedAccount
) -> Optional[DivaPoolParameters]:
    """Fetches and parses needed parameters for a given pool."""

    diva = DivaProtocolContract(node, account)
    diva.connect()

    params = await diva.get_pool_parameters(pool_id)
    if params is None:
        logger.error("Error getting pool params from Diva contract.")
        return None

    print("PARAMS:", params)
    pool_params = DivaPoolParameters(
        reference_asset=params.reference_asset, expiry_date=params.expiry_time
    )
    if pool_params.reference_asset not in SUPPORTED_REFERENCE_ASSETS:
        logger.error(f"Reference asset not supported: {pool_params.reference_asset}")
        return None

    return pool_params


def get_source(asset: str, ts: int) -> PriceAggregator:
    """Returns PriceAggregator with sources adjusted based on given asset."""
    source = PriceAggregator(
        asset=asset,
        currency="usd",
        algorithm="median",
        sources=[
            CryptowatchHistoricalPriceSource(asset=asset, currency="usd", ts=ts),
            KrakenHistoricalPriceSource(asset=asset, currency="usd", ts=ts),
            PoloniexHistoricalPriceSource(asset=asset, currency="dai", ts=ts),
            PoloniexHistoricalPriceSource(asset=asset, currency="tusd", ts=ts),
        ],
    )
    if asset == "btc":
        source.sources[1].asset = "xbt"

    return source


async def assemble_diva_datafeed(
    pool_id: int, node: RPCEndpoint, account: ChainedAccount
) -> Optional[DataFeed[float]]:
    """Returns datafeed using user input pool ID and corresponding
    asset information.

    Reference assets are currently whitelisted & hard-coded."""

    params = await get_pool_params(pool_id, node, account)
    if params is None:
        logger.error("Error getting pool parameters.")
        return None

    asset = params.reference_asset.split("/")[0].lower()
    ts = params.expiry_date

    feed = DataFeed(
        query=divaProtocolPolygon(pool_id),
        source=get_source(asset, ts),
    )

    return feed



"""Datafeed for current price of ETH in USD used by LegacyQueryReporter."""
from telliot_core.datafeed import DataFeed
from telliot_core.queries import LegacyRequest

from telliot_feed_examples.sources.price.spot.binance import BinanceSpotPriceSource
from telliot_feed_examples.sources.price.spot.coinbase import CoinbaseSpotPriceSource
from telliot_feed_examples.sources.price.spot.coingecko import CoinGeckoSpotPriceSource
from telliot_feed_examples.sources.price.spot.kraken import KrakenSpotPriceSource
from telliot_feed_examples.sources.price_aggregator import PriceAggregator

eth_usd_median_feed = DataFeed(
    query=LegacyRequest(legacy_id=1),
    source=PriceAggregator(
        asset="eth",
        currency="usd",
        algorithm="median",
        sources=[
            CoinbaseSpotPriceSource(asset="eth", currency="usd"),
            CoinGeckoSpotPriceSource(asset="eth", currency="usd"),
            KrakenSpotPriceSource(asset="eth", currency="usd"),
            BinanceSpotPriceSource(asset="eth", currency="usdc"),
            BinanceSpotPriceSource(asset="eth", currency="usdt"),
        ],
    ),
)



import yaml

from telliot_feed_examples.feeds.ampl_usd_vwap_feed import ampl_usd_vwap_feed


def test_ampl_feed():
    """Test AMPL/USD/VWAP feed."""
    feed = ampl_usd_vwap_feed

    print(yaml.dump(feed.get_state(), sort_keys=False))



from telliot_feed_examples.feeds.ampl_usd_vwap_feed import ampl_usd_vwap_feed
from telliot_feed_examples.feeds.bct_usd_feed import bct_usd_median_feed
from telliot_feed_examples.feeds.btc_usd_feed import btc_usd_median_feed
from telliot_feed_examples.feeds.dai_usd_feed import dai_usd_median_feed
from telliot_feed_examples.feeds.eth_jpy_feed import eth_jpy_median_feed
from telliot_feed_examples.feeds.eth_usd_feed import eth_usd_median_feed
from telliot_feed_examples.feeds.idle_usd_feed import idle_usd_median_feed
from telliot_feed_examples.feeds.matic_usd_feed import matic_usd_median_feed
from telliot_feed_examples.feeds.mkr_usd_feed import mkr_usd_median_feed
from telliot_feed_examples.feeds.olympus import ohm_eth_median_feed
from telliot_feed_examples.feeds.ric_usd_feed import ric_usd_median_feed
from telliot_feed_examples.feeds.sushi_usd_feed import sushi_usd_median_feed
from telliot_feed_examples.feeds.trb_usd_feed import trb_usd_median_feed
from telliot_feed_examples.feeds.usdc_usd_feed import usdc_usd_median_feed
from telliot_feed_examples.feeds.uspce_feed import uspce_feed
from telliot_feed_examples.feeds.vesq import vsq_usd_median_feed


# Supported legacy feeds
LEGACY_DATAFEEDS = {
    "1": eth_usd_median_feed,
    "2": btc_usd_median_feed,
    "10": ampl_usd_vwap_feed,
    "41": uspce_feed,
    "50": trb_usd_median_feed,
    "59": eth_jpy_median_feed,
}

CATALOG_FEEDS = {
    "eth-usd-legacy": eth_usd_median_feed,
    "btc-usd-legacy": btc_usd_median_feed,
    "ampl-legacy": ampl_usd_vwap_feed,
    "uspce-legacy": uspce_feed,
    "trb-usd-legacy": trb_usd_median_feed,
    "eth-jpy-legacy": eth_jpy_median_feed,
    "ohm-eth-spot": ohm_eth_median_feed,
    "vsq-usd-spot": vsq_usd_median_feed,
    "bct-usd-spot": bct_usd_median_feed,
    "dai-usd-spot": dai_usd_median_feed,
    "ric-usd-spot": ric_usd_median_feed,
    "idle-usd-spot": idle_usd_median_feed,
    "mkr-usd-spot": mkr_usd_median_feed,
    "sushi-usd-spot": sushi_usd_median_feed,
    "matic-usd-spot": matic_usd_median_feed,
    "usdc-usd-spot": usdc_usd_median_feed,
}



import statistics

import pytest

from telliot_feed_examples.feeds.eth_usd_feed import eth_usd_median_feed


@pytest.mark.asyncio
async def test_AssetPriceFeed():
    """Retrieve median ETH/USD price."""
    v, _ = await eth_usd_median_feed.source.fetch_new_datapoint()

    assert v is not None
    assert v > 0
    print(f"ETH/USD Price: {v}")

    # Get list of data sources from sources dict
    source_prices = [source.latest[0] for source in eth_usd_median_feed.source.sources]

    # Make sure error is less than decimal tolerance
    assert (v - statistics.median(source_prices)) < 10**-6



from dataclasses import dataclass
from dataclasses import field
from typing import Any

import requests
from telliot_core.dtypes.datapoint import datetime_now_utc
from telliot_core.dtypes.datapoint import OptionalDataPoint
from telliot_core.pricing.price_service import WebPriceService
from telliot_core.pricing.price_source import PriceSource

from telliot_feed_examples.utils.log import get_logger


logger = get_logger(__name__)
uniswapV3_map = {
    "eth": "0x0000000000000000000000000000000000000000",
    "wbtc": "0x2260fac5e5542a773aa44fbcfedf7c193bc2c599",
    "matic": "0x7d1afa7b718fb893db30a3abc0cfc608aacfebb0",
    "dai": "0x6b175474e89094c44da98b954eedeac495271d0f",
    "fuse": "0x970b9bb2c0444f5e81e9d0efb84c8ccdcdcaf84d",
}


class UniswapV3PriceService(WebPriceService):
    """UniswapV3 Price Service in USD and ETH"""

    def __init__(self, **kwargs: Any) -> None:
        kwargs["name"] = "UniswapV3 Price Service"
        kwargs["url"] = "https://api.thegraph.com"
        kwargs["timeout"] = 10.0
        super().__init__(**kwargs)

    async def get_price(self, asset: str, currency: str) -> OptionalDataPoint[float]:
        """Implement PriceServiceInterface

        This implementation gets the price from the UniswapV3 subgraph
        https://docs.uniswap.org/sdk/subgraph/subgraph-examples

        """

        asset = asset.lower()

        token = uniswapV3_map.get(asset, None)
        if not token:
            raise Exception("Asset not supported: {}".format(asset))

        headers = {
            "Content-Type": "application/json",
        }

        query = (
            "{bundles{id ethPriceUSD}token" + f'(id: "{token}")' + "{ derivedETH } }"
        )

        json_data = {"query": query}

        request_url = self.url + "/subgraphs/name/uniswap/uniswap-v3"

        with requests.Session() as s:
            try:
                r = s.post(
                    request_url, headers=headers, json=json_data, timeout=self.timeout
                )
                res = r.json()
                data = {"response": res}

            except requests.exceptions.ConnectTimeout:
                logger.warning("Timeout Error, No prices retrieved from Uniswap")
                return None, None

            except Exception:
                logger.warning("No prices retrieved from Uniswap")
                return None, None

        if "error" in data:
            logger.error(data)
            return None, None

        elif "response" in data:
            response = data["response"]

            try:
                ethprice = float(response["data"]["bundles"][0]["ethPriceUSD"])
                if asset.lower() == "eth":
                    token_data = 1
                elif currency.lower() == "eth":
                    ethprice = 1
                    token_data = response["data"]["token"]["derivedETH"]
                else:
                    token_data = response["data"]["token"]["derivedETH"]
                price = ethprice * float(token_data)
                return price, datetime_now_utc()
            except KeyError as e:
                msg = "Error parsing UniswapV3 response: KeyError: {}".format(e)
                logger.critical(msg)
                return None, None

        else:
            raise Exception("Invalid response from get_url")


@dataclass
class UniswapV3PriceSource(PriceSource):
    asset: str = ""
    currency: str = ""
    service: UniswapV3PriceService = field(
        default_factory=UniswapV3PriceService, init=False
    )



"""Datafeed for current price of ETH in JPY used by LegacyQueryReporter."""
from telliot_core.datafeed import DataFeed
from telliot_core.queries import LegacyRequest

from telliot_feed_examples.sources.price.spot.bitfinex import BitfinexSpotPriceSource
from telliot_feed_examples.sources.price.spot.bitflyer import BitflyerSpotPriceSource
from telliot_feed_examples.sources.price.spot.coingecko import CoinGeckoSpotPriceSource
from telliot_feed_examples.sources.price_aggregator import PriceAggregator

eth_jpy_median_feed = DataFeed(
    query=LegacyRequest(legacy_id=59),
    source=PriceAggregator(
        asset="eth",
        currency="jpy",
        algorithm="median",
        sources=[
            CoinGeckoSpotPriceSource(asset="eth", currency="jpy"),
            BitflyerSpotPriceSource(asset="eth", currency="jpy"),
            BitfinexSpotPriceSource(asset="eth", currency="jpy"),
        ],
    ),
)



""" Unit tests for pricing module

"""
import os
from datetime import datetime

import pytest
from telliot_core.apps.telliot_config import TelliotConfig

from telliot_feed_examples.sources.price.spot.bittrex import BittrexSpotPriceService
from telliot_feed_examples.sources.price.spot.coinbase import CoinbaseSpotPriceService
from telliot_feed_examples.sources.price.spot.coingecko import CoinGeckoSpotPriceService
from telliot_feed_examples.sources.price.spot.gemini import GeminiSpotPriceService
from telliot_feed_examples.sources.price.spot.nomics import NomicsSpotPriceService
from telliot_feed_examples.sources.price.spot.pancakeswap import (
    PancakeswapPriceService,
)
from telliot_feed_examples.sources.price.spot.uniswapV3 import UniswapV3PriceService


service = {
    "coinbase": CoinbaseSpotPriceService(),
    "coingecko": CoinGeckoSpotPriceService(),
    "bittrex": BittrexSpotPriceService(),
    "gemini": GeminiSpotPriceService(),
    "nomics": NomicsSpotPriceService(),
    "pancakeswap": PancakeswapPriceService(),
    "uniswapV3": UniswapV3PriceService(),
}


async def get_price(asset, currency, s, timeout=10.0):
    """Helper function for retrieving prices."""
    s.timeout = timeout
    v, t = await s.get_price(asset, currency)
    return v, t


def validate_price(v, t):
    """Check types and price anomalies."""
    assert v is not None
    assert isinstance(v, float)
    assert v > 0
    assert isinstance(t, datetime)
    print(v)
    print(t)


@pytest.fixture()
def nomics_key():
    key = TelliotConfig().api_keys.find(name="nomics")[0].key

    if not key and "NOMICS_KEY" in os.environ:
        key = os.environ["NOMICS_KEY"]

    return key


@pytest.fixture()
def coinmarketcap_key():
    key = TelliotConfig().api_keys.find(name="coinmarketcap")[0].key

    if not key and "COINMARKETCAP_KEY" in os.environ:
        key = os.environ["COINMARKETCAP_KEY"]

    return key


@pytest.mark.asyncio
async def test_coinbase():
    """Test retrieving from Coinbase price source."""
    v, t = await get_price("btc", "usd", service["coinbase"])
    validate_price(v, t)


@pytest.mark.asyncio
async def test_coingecko():
    """Test retrieving from Coingecko price source."""
    v, t = await get_price("btc", "usd", service["coingecko"])
    validate_price(v, t)


@pytest.mark.asyncio
async def test_nomics(nomics_key):
    """Test retrieving from Nomics price source."""
    if nomics_key:
        v, t = await get_price("btc", "usd", service["nomics"])
        validate_price(v, t)
    else:
        print("No Nomics API key ")


@pytest.mark.asyncio
async def test_coinmarketcap(coinmarketcap_key):
    """Test retrieving from CoinMarketCap price source."""
    if coinmarketcap_key:
        v, t = await get_price("bct", "usd", service["coinmarketcap"])
        validate_price(v, t)
    else:
        print("No CoinMarketCap API key ")


@pytest.mark.asyncio
async def test_bittrex():
    """Test retrieving from Bittrex price source."""
    v, t = await get_price("btc", "usd", service["bittrex"])
    validate_price(v, t)


@pytest.mark.asyncio
async def test_gemini():
    """Test retrieving from Gemini price source."""
    v, t = await get_price("btc", "usd", service["gemini"])
    validate_price(v, t)


@pytest.mark.asyncio
async def test_uniswap_usd():
    """Test retrieving from UniswapV3 price source in USD."""
    v, t = await get_price("fuse", "usd", service["uniswapV3"])
    validate_price(v, t)


@pytest.mark.asyncio
async def test_uniswap_timeout():
    """Test retrieving from UniswapV3 price source in USD."""
    v, t = await get_price("fuse", "usd", service["uniswapV3"], 0.05)
    assert v is None
    assert t is None


@pytest.mark.asyncio
async def test_uniswap_eth():
    """Test retrieving from UniswapV3 price source in ETH."""
    v, t = await get_price("fuse", "eth", service["uniswapV3"])
    validate_price(v, t)


@pytest.mark.asyncio
async def test_uniswap_eth_usd():
    """Test retrieving from UniswapV3 price source for Eth in USD."""
    v, t = await get_price("eth", "usd", service["uniswapV3"])
    validate_price(v, t)


@pytest.mark.asyncio
async def test_pancakeswap_usd():
    """Test retrieving from Pancakeswap price source in USD."""
    v, t = await get_price("fuse", "usd", service["pancakeswap"])
    validate_price(v, t)


@pytest.mark.asyncio
async def test_pancakeswap_bnb():
    """Test retrieving from Pancakeswap price source in BNB."""
    v, t = await get_price("fuse", "bnb", service["pancakeswap"])
    validate_price(v, t)


# def test_web_price_service_timeout():
#     ps = CoinbaseSpotPriceService(timeout=0.0000001)
#     result = ps.get_url()
#     assert result["error"] == "Timeout Error"



from dataclasses import dataclass
from dataclasses import field
from typing import Any
from urllib.parse import urlencode

from telliot_core.dtypes.datapoint import datetime_now_utc
from telliot_core.dtypes.datapoint import OptionalDataPoint
from telliot_core.pricing.price_service import WebPriceService
from telliot_core.pricing.price_source import PriceSource

from telliot_feed_examples.utils.log import get_logger


logger = get_logger(__name__)


# Hardcoded supported assets & currencies
bitflyer_assets = {"ETH"}
bitflyer_currencies = {"JPY"}


class BitflyerSpotPriceService(WebPriceService):
    """Bitflyer Price Service"""

    def __init__(self, **kwargs: Any) -> None:
        kwargs["name"] = "Bitflyer Price Service"
        kwargs["url"] = "https://api.bitflyer.com"
        super().__init__(**kwargs)

    async def get_price(self, asset: str, currency: str) -> OptionalDataPoint[float]:
        """Implement PriceServiceInterface

        This implementation gets the price from the Bitflyer API."""

        asset = asset.upper()
        currency = currency.upper()

        if asset not in bitflyer_assets:
            raise Exception(f"Asset not supported: {asset}")
        if currency not in bitflyer_currencies:
            raise Exception(f"Currency not supported: {currency}")

        asset_currency = asset + "_" + currency

        url_params = urlencode({"product_code": asset_currency})
        request_url = f"/v1/getticker?{url_params}"

        d = self.get_url(request_url)

        if "error" in d:
            logger.error(d)
            return None, None

        elif "response" in d:
            response = d["response"]

            try:
                price = float(response["ltp"])
            except KeyError as e:
                msg = f"Error parsing Coingecko API response: KeyError: {e}"
                logger.critical(msg)

        else:
            raise Exception("Invalid response from get_url")

        return price, datetime_now_utc()


@dataclass
class BitflyerSpotPriceSource(PriceSource):
    asset: str = ""
    currency: str = ""
    service: BitflyerSpotPriceService = field(
        default_factory=BitflyerSpotPriceService, init=False
    )



import pytest
from telliot_core.apps.core import TelliotCore
from web3.datastructures import AttributeDict

from telliot_feed_examples.feeds.dai_usd_feed import dai_usd_median_feed
from telliot_feed_examples.reporters.tellorflex import PolygonReporter


@pytest.mark.skip("Run locally, github action testing failure")
@pytest.mark.asyncio
async def test_dai_usd_reporter_submit_once(mumbai_cfg):
    """Test reporting bct/usd on mumbai."""
    async with TelliotCore(config=mumbai_cfg) as core:
        flex = core.get_tellorflex_contracts()
        r = PolygonReporter(
            endpoint=core.endpoint,
            account=core.get_account(),
            chain_id=80001,
            oracle=flex.oracle,
            token=flex.token,
            datafeed=dai_usd_median_feed,
            max_fee=100,
        )

        ORACLE_ADDRESSES = {
            "0xFd45Ae72E81Adaaf01cC61c8bCe016b7060DD537",  # polygon
            "0x41b66dd93b03e89D29114a7613A6f9f0d4F40178",  # mumbai
        }

        tx_receipt, status = await r.report_once()

        # Reporter submitted
        if tx_receipt is not None and status.ok:
            assert isinstance(tx_receipt, AttributeDict)
            assert tx_receipt.to in ORACLE_ADDRESSES
        # Reporter did not submit
        else:
            assert not tx_receipt
            assert not status.ok
            assert (
                ("Currently in reporter lock." in status.error)
                or ("Current addess disputed" in status.error)
                or ("Unable to retrieve updated datafeed" in status.error)
            )



import pytest
from telliot_core.apps.core import TelliotCore
from telliot_core.utils.response import ResponseStatus

from telliot_feed_examples.feeds.eth_usd_feed import eth_usd_median_feed
from telliot_feed_examples.reporters.tellorflex import PolygonReporter


@pytest.mark.skip("mumbai cfg error in github actions, but not locally")
@pytest.fixture
async def polygon_reporter(mumbai_cfg):
    async with TelliotCore(config=mumbai_cfg) as core:
        flex = core.get_tellorflex_contracts()
        r = PolygonReporter(
            oracle=flex.oracle,
            token=flex.token,
            endpoint=core.endpoint,
            account=core.get_account(),
            chain_id=80001,
        )
        return r


@pytest.mark.skip("mumbai cfg error in github actions, but not locally")
@pytest.mark.asyncio
async def test_ensure_profitable(polygon_reporter):
    status = await polygon_reporter.ensure_profitable(eth_usd_median_feed)

    assert isinstance(status, ResponseStatus)
    assert status.ok


@pytest.mark.skip("mumbai cfg error in github actions, but not locally")
@pytest.mark.asyncio
async def test_fetch_gas_price(polygon_reporter):
    price = await polygon_reporter.fetch_gas_price()

    assert isinstance(price, int)
    assert price > 0


@pytest.mark.skip("mumbai cfg error in github actions, but not locally")
@pytest.mark.asyncio
async def test_ensure_staked(polygon_reporter):
    staked, status = await polygon_reporter.ensure_staked()

    assert isinstance(status, ResponseStatus)
    assert isinstance(staked, bool)
    if status.ok:
        assert staked
    else:
        assert "Unable to approve staking" in status.error


@pytest.mark.skip("mumbai cfg error in github actions, but not locally")
@pytest.mark.asyncio
async def test_check_reporter_lock(polygon_reporter):
    status = await polygon_reporter.check_reporter_lock()

    assert isinstance(status, ResponseStatus)
    if not status.ok:
        assert ("reporter lock" in status.error) or (
            "Staker balance too low" in status.error
        )


@pytest.mark.skip("mumbai cfg error in github actions, but not locally")
@pytest.mark.asyncio
async def test_get_num_reports_by_id(polygon_reporter):
    qid = eth_usd_median_feed.query.query_id
    count, status = await polygon_reporter.get_num_reports_by_id(qid)

    assert isinstance(status, ResponseStatus)
    if status.ok:
        assert isinstance(count, int)
    else:
        assert count is None



from datetime import datetime

import pytest

from telliot_feed_examples.sources import diva_protocol
from telliot_feed_examples.sources.diva_protocol import DivaManualSource


@pytest.mark.asyncio
async def test_uspce_source():
    """Test retrieving USPCE data from user input."""
    # Override Python built-in input method
    diva_protocol.input = lambda: "1234.5678"

    source = DivaManualSource(reference_asset="BTC/USD", timestamp=0)

    val, dt = await source.fetch_new_datapoint()

    assert isinstance(dt, datetime)
    assert dt.year == 1970
    assert val == 1234.5678



from telliot_core.data.query_catalog import query_catalog

from telliot_feed_examples.feeds import CATALOG_FEEDS


def test_supports_all_active_queries():
    """Make sure all current queries have an associated feed that reporters can use."""
    active_q_tags = [q.tag for q in query_catalog.find()]

    for qt in active_q_tags:

        # Skip AWS spot price bc haven't not sure needed for morphware integration
        if "aws-spot-zone" in qt:
            continue

        assert qt in CATALOG_FEEDS
